# coding: utf-8

# flake8: noqa

"""
    Kubeflow Training SDK

    Python SDK for Kubeflow Training  # noqa: E501

    The version of the OpenAPI document: v1.9.2
    Generated by: https://openapi-generator.tech
"""


from __future__ import absolute_import

__version__ = "1.9.2"

# import apis into sdk package

# import ApiClient
from kubeflow.training.api_client import ApiClient
from kubeflow.training.configuration import Configuration
from kubeflow.training.exceptions import OpenApiException
from kubeflow.training.exceptions import ApiTypeError
from kubeflow.training.exceptions import ApiValueError
from kubeflow.training.exceptions import ApiKeyError
from kubeflow.training.exceptions import ApiException
# import models into sdk package
from kubeflow.training.models.kubeflow_org_v1_elastic_policy import KubeflowOrgV1ElasticPolicy
from kubeflow.training.models.kubeflow_org_v1_jax_job import KubeflowOrgV1JAXJob
from kubeflow.training.models.kubeflow_org_v1_jax_job_list import KubeflowOrgV1JAXJobList
from kubeflow.training.models.kubeflow_org_v1_jax_job_spec import KubeflowOrgV1JAXJobSpec
from kubeflow.training.models.kubeflow_org_v1_job_condition import KubeflowOrgV1JobCondition
from kubeflow.training.models.kubeflow_org_v1_job_status import KubeflowOrgV1JobStatus
from kubeflow.training.models.kubeflow_org_v1_mpi_job import KubeflowOrgV1MPIJob
from kubeflow.training.models.kubeflow_org_v1_mpi_job_list import KubeflowOrgV1MPIJobList
from kubeflow.training.models.kubeflow_org_v1_mpi_job_spec import KubeflowOrgV1MPIJobSpec
from kubeflow.training.models.kubeflow_org_v1_paddle_elastic_policy import KubeflowOrgV1PaddleElasticPolicy
from kubeflow.training.models.kubeflow_org_v1_paddle_job import KubeflowOrgV1PaddleJob
from kubeflow.training.models.kubeflow_org_v1_paddle_job_list import KubeflowOrgV1PaddleJobList
from kubeflow.training.models.kubeflow_org_v1_paddle_job_spec import KubeflowOrgV1PaddleJobSpec
from kubeflow.training.models.kubeflow_org_v1_py_torch_job import KubeflowOrgV1PyTorchJob
from kubeflow.training.models.kubeflow_org_v1_py_torch_job_list import KubeflowOrgV1PyTorchJobList
from kubeflow.training.models.kubeflow_org_v1_py_torch_job_spec import KubeflowOrgV1PyTorchJobSpec
from kubeflow.training.models.kubeflow_org_v1_rdzv_conf import KubeflowOrgV1RDZVConf
from kubeflow.training.models.kubeflow_org_v1_replica_spec import KubeflowOrgV1ReplicaSpec
from kubeflow.training.models.kubeflow_org_v1_replica_status import KubeflowOrgV1ReplicaStatus
from kubeflow.training.models.kubeflow_org_v1_run_policy import KubeflowOrgV1RunPolicy
from kubeflow.training.models.kubeflow_org_v1_scheduling_policy import KubeflowOrgV1SchedulingPolicy
from kubeflow.training.models.kubeflow_org_v1_tf_job import KubeflowOrgV1TFJob
from kubeflow.training.models.kubeflow_org_v1_tf_job_list import KubeflowOrgV1TFJobList
from kubeflow.training.models.kubeflow_org_v1_tf_job_spec import KubeflowOrgV1TFJobSpec
from kubeflow.training.models.kubeflow_org_v1_xg_boost_job import KubeflowOrgV1XGBoostJob
from kubeflow.training.models.kubeflow_org_v1_xg_boost_job_list import KubeflowOrgV1XGBoostJobList
from kubeflow.training.models.kubeflow_org_v1_xg_boost_job_spec import KubeflowOrgV1XGBoostJobSpec

from kubeflow.training.api.training_client import TrainingClient
from kubeflow.training.constants import constants



import argparse

from .hugging_face import HuggingFace, HuggingFaceDataset
from .s3 import S3


def model_factory(model_provider, model_provider_parameters):
    match model_provider:
        case "hf":
            hf = HuggingFace()
            hf.load_config(model_provider_parameters)
            hf.download_model_and_tokenizer()
        case _:
            return "This is the default case"


def dataset_factory(dataset_provider, dataset_provider_parameters):
    match dataset_provider:
        case "s3":
            s3 = S3()
            s3.load_config(dataset_provider_parameters)
            s3.download_dataset()
        case "hf":
            hf = HuggingFaceDataset()
            hf.load_config(dataset_provider_parameters)
            hf.download_dataset()
        case _:
            return "This is the default case"


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="script for downloading model and datasets to PVC."
    )
    parser.add_argument("--model_provider", type=str, help="name of model provider")
    parser.add_argument(
        "--model_provider_parameters",
        type=str,
        help="model provider serialised arguments",
    )

    parser.add_argument("--dataset_provider", type=str, help="name of dataset provider")
    parser.add_argument(
        "--dataset_provider_parameters",
        type=str,
        help="dataset provider serialized arguments",
    )
    args = parser.parse_args()

    model_factory(args.model_provider, args.model_provider_parameters)
    dataset_factory(args.dataset_provider, args.dataset_provider_parameters)



# coding: utf-8

"""
    Kubeflow Training SDK

    Python SDK for Kubeflow Training  # noqa: E501

    The version of the OpenAPI document: v1.9.2
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from kubeflow.training.configuration import Configuration


class KubeflowOrgV1JobStatus(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'completion_time': 'datetime',
        'conditions': 'list[KubeflowOrgV1JobCondition]',
        'last_reconcile_time': 'datetime',
        'replica_statuses': 'dict(str, KubeflowOrgV1ReplicaStatus)',
        'start_time': 'datetime'
    }

    attribute_map = {
        'completion_time': 'completionTime',
        'conditions': 'conditions',
        'last_reconcile_time': 'lastReconcileTime',
        'replica_statuses': 'replicaStatuses',
        'start_time': 'startTime'
    }

    def __init__(self, completion_time=None, conditions=None, last_reconcile_time=None, replica_statuses=None, start_time=None, local_vars_configuration=None):  # noqa: E501
        """KubeflowOrgV1JobStatus - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._completion_time = None
        self._conditions = None
        self._last_reconcile_time = None
        self._replica_statuses = None
        self._start_time = None
        self.discriminator = None

        if completion_time is not None:
            self.completion_time = completion_time
        if conditions is not None:
            self.conditions = conditions
        if last_reconcile_time is not None:
            self.last_reconcile_time = last_reconcile_time
        if replica_statuses is not None:
            self.replica_statuses = replica_statuses
        if start_time is not None:
            self.start_time = start_time

    @property
    def completion_time(self):
        """Gets the completion_time of this KubeflowOrgV1JobStatus.  # noqa: E501


        :return: The completion_time of this KubeflowOrgV1JobStatus.  # noqa: E501
        :rtype: datetime
        """
        return self._completion_time

    @completion_time.setter
    def completion_time(self, completion_time):
        """Sets the completion_time of this KubeflowOrgV1JobStatus.


        :param completion_time: The completion_time of this KubeflowOrgV1JobStatus.  # noqa: E501
        :type: datetime
        """

        self._completion_time = completion_time

    @property
    def conditions(self):
        """Gets the conditions of this KubeflowOrgV1JobStatus.  # noqa: E501

        Conditions is an array of current observed job conditions.  # noqa: E501

        :return: The conditions of this KubeflowOrgV1JobStatus.  # noqa: E501
        :rtype: list[KubeflowOrgV1JobCondition]
        """
        return self._conditions

    @conditions.setter
    def conditions(self, conditions):
        """Sets the conditions of this KubeflowOrgV1JobStatus.

        Conditions is an array of current observed job conditions.  # noqa: E501

        :param conditions: The conditions of this KubeflowOrgV1JobStatus.  # noqa: E501
        :type: list[KubeflowOrgV1JobCondition]
        """

        self._conditions = conditions

    @property
    def last_reconcile_time(self):
        """Gets the last_reconcile_time of this KubeflowOrgV1JobStatus.  # noqa: E501


        :return: The last_reconcile_time of this KubeflowOrgV1JobStatus.  # noqa: E501
        :rtype: datetime
        """
        return self._last_reconcile_time

    @last_reconcile_time.setter
    def last_reconcile_time(self, last_reconcile_time):
        """Sets the last_reconcile_time of this KubeflowOrgV1JobStatus.


        :param last_reconcile_time: The last_reconcile_time of this KubeflowOrgV1JobStatus.  # noqa: E501
        :type: datetime
        """

        self._last_reconcile_time = last_reconcile_time

    @property
    def replica_statuses(self):
        """Gets the replica_statuses of this KubeflowOrgV1JobStatus.  # noqa: E501

        ReplicaStatuses is map of ReplicaType and ReplicaStatus, specifies the status of each replica.  # noqa: E501

        :return: The replica_statuses of this KubeflowOrgV1JobStatus.  # noqa: E501
        :rtype: dict(str, KubeflowOrgV1ReplicaStatus)
        """
        return self._replica_statuses

    @replica_statuses.setter
    def replica_statuses(self, replica_statuses):
        """Sets the replica_statuses of this KubeflowOrgV1JobStatus.

        ReplicaStatuses is map of ReplicaType and ReplicaStatus, specifies the status of each replica.  # noqa: E501

        :param replica_statuses: The replica_statuses of this KubeflowOrgV1JobStatus.  # noqa: E501
        :type: dict(str, KubeflowOrgV1ReplicaStatus)
        """

        self._replica_statuses = replica_statuses

    @property
    def start_time(self):
        """Gets the start_time of this KubeflowOrgV1JobStatus.  # noqa: E501


        :return: The start_time of this KubeflowOrgV1JobStatus.  # noqa: E501
        :rtype: datetime
        """
        return self._start_time

    @start_time.setter
    def start_time(self, start_time):
        """Sets the start_time of this KubeflowOrgV1JobStatus.


        :param start_time: The start_time of this KubeflowOrgV1JobStatus.  # noqa: E501
        :type: datetime
        """

        self._start_time = start_time

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, KubeflowOrgV1JobStatus):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, KubeflowOrgV1JobStatus):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Kubeflow Training SDK

    Python SDK for Kubeflow Training  # noqa: E501

    The version of the OpenAPI document: v1.9.2
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from kubeflow.training.configuration import Configuration


class KubeflowOrgV1SchedulingPolicy(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'min_available': 'int',
        'min_resources': 'dict(str, Quantity)',
        'priority_class': 'str',
        'queue': 'str',
        'schedule_timeout_seconds': 'int'
    }

    attribute_map = {
        'min_available': 'minAvailable',
        'min_resources': 'minResources',
        'priority_class': 'priorityClass',
        'queue': 'queue',
        'schedule_timeout_seconds': 'scheduleTimeoutSeconds'
    }

    def __init__(self, min_available=None, min_resources=None, priority_class=None, queue=None, schedule_timeout_seconds=None, local_vars_configuration=None):  # noqa: E501
        """KubeflowOrgV1SchedulingPolicy - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._min_available = None
        self._min_resources = None
        self._priority_class = None
        self._queue = None
        self._schedule_timeout_seconds = None
        self.discriminator = None

        if min_available is not None:
            self.min_available = min_available
        if min_resources is not None:
            self.min_resources = min_resources
        if priority_class is not None:
            self.priority_class = priority_class
        if queue is not None:
            self.queue = queue
        if schedule_timeout_seconds is not None:
            self.schedule_timeout_seconds = schedule_timeout_seconds

    @property
    def min_available(self):
        """Gets the min_available of this KubeflowOrgV1SchedulingPolicy.  # noqa: E501


        :return: The min_available of this KubeflowOrgV1SchedulingPolicy.  # noqa: E501
        :rtype: int
        """
        return self._min_available

    @min_available.setter
    def min_available(self, min_available):
        """Sets the min_available of this KubeflowOrgV1SchedulingPolicy.


        :param min_available: The min_available of this KubeflowOrgV1SchedulingPolicy.  # noqa: E501
        :type: int
        """

        self._min_available = min_available

    @property
    def min_resources(self):
        """Gets the min_resources of this KubeflowOrgV1SchedulingPolicy.  # noqa: E501


        :return: The min_resources of this KubeflowOrgV1SchedulingPolicy.  # noqa: E501
        :rtype: dict(str, Quantity)
        """
        return self._min_resources

    @min_resources.setter
    def min_resources(self, min_resources):
        """Sets the min_resources of this KubeflowOrgV1SchedulingPolicy.


        :param min_resources: The min_resources of this KubeflowOrgV1SchedulingPolicy.  # noqa: E501
        :type: dict(str, Quantity)
        """

        self._min_resources = min_resources

    @property
    def priority_class(self):
        """Gets the priority_class of this KubeflowOrgV1SchedulingPolicy.  # noqa: E501


        :return: The priority_class of this KubeflowOrgV1SchedulingPolicy.  # noqa: E501
        :rtype: str
        """
        return self._priority_class

    @priority_class.setter
    def priority_class(self, priority_class):
        """Sets the priority_class of this KubeflowOrgV1SchedulingPolicy.


        :param priority_class: The priority_class of this KubeflowOrgV1SchedulingPolicy.  # noqa: E501
        :type: str
        """

        self._priority_class = priority_class

    @property
    def queue(self):
        """Gets the queue of this KubeflowOrgV1SchedulingPolicy.  # noqa: E501


        :return: The queue of this KubeflowOrgV1SchedulingPolicy.  # noqa: E501
        :rtype: str
        """
        return self._queue

    @queue.setter
    def queue(self, queue):
        """Sets the queue of this KubeflowOrgV1SchedulingPolicy.


        :param queue: The queue of this KubeflowOrgV1SchedulingPolicy.  # noqa: E501
        :type: str
        """

        self._queue = queue

    @property
    def schedule_timeout_seconds(self):
        """Gets the schedule_timeout_seconds of this KubeflowOrgV1SchedulingPolicy.  # noqa: E501


        :return: The schedule_timeout_seconds of this KubeflowOrgV1SchedulingPolicy.  # noqa: E501
        :rtype: int
        """
        return self._schedule_timeout_seconds

    @schedule_timeout_seconds.setter
    def schedule_timeout_seconds(self, schedule_timeout_seconds):
        """Sets the schedule_timeout_seconds of this KubeflowOrgV1SchedulingPolicy.


        :param schedule_timeout_seconds: The schedule_timeout_seconds of this KubeflowOrgV1SchedulingPolicy.  # noqa: E501
        :type: int
        """

        self._schedule_timeout_seconds = schedule_timeout_seconds

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, KubeflowOrgV1SchedulingPolicy):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, KubeflowOrgV1SchedulingPolicy):
            return True

        return self.to_dict() != other.to_dict()



import json
import logging
from dataclasses import dataclass, field
from typing import Optional, Union
from urllib.parse import urlparse

import transformers
from peft import LoraConfig

from .abstract_dataset_provider import datasetProvider
from .abstract_model_provider import modelProvider
from .constants import VOLUME_PATH_DATASET, VOLUME_PATH_MODEL

TRANSFORMER_TYPES = Union[
    transformers.AutoModelForSequenceClassification,
    transformers.AutoModelForTokenClassification,
    transformers.AutoModelForQuestionAnswering,
    transformers.AutoModelForCausalLM,
    transformers.AutoModelForMaskedLM,
    transformers.AutoModelForImageClassification,
]


# Configure logger.
log_formatter = logging.Formatter(
    "%(asctime)s %(levelname)-8s %(message)s", "%Y-%m-%dT%H:%M:%SZ"
)
logger = logging.getLogger(__file__)
console_handler = logging.StreamHandler()
console_handler.setFormatter(log_formatter)
logger.addHandler(console_handler)
logger.setLevel(logging.INFO)


@dataclass
class HuggingFaceModelParams:
    model_uri: str
    transformer_type: TRANSFORMER_TYPES
    access_token: str = None
    num_labels: Optional[int] = None

    def __post_init__(self):
        # Custom checks or validations can be added here
        if self.model_uri == "" or self.model_uri is None:
            raise ValueError("model_uri cannot be empty.")


@dataclass
class HuggingFaceTrainerParams:
    training_parameters: transformers.TrainingArguments = field(
        default_factory=transformers.TrainingArguments
    )
    lora_config: LoraConfig = field(default_factory=LoraConfig)


class HuggingFace(modelProvider):
    def load_config(self, serialised_args):
        # implementation for loading the config
        self.config = HuggingFaceModelParams(**json.loads(serialised_args))

    def download_model_and_tokenizer(self):
        # implementation for downloading the model
        logger.info("Downloading model")
        logger.info("-" * 40)
        transformer_type_class = getattr(transformers, self.config.transformer_type)
        parsed_uri = urlparse(self.config.model_uri)
        self.model = parsed_uri.netloc + parsed_uri.path
        transformer_type_class.from_pretrained(
            self.model,
            token=self.config.access_token,
            cache_dir=VOLUME_PATH_MODEL,
            trust_remote_code=True,
        )
        transformers.AutoTokenizer.from_pretrained(
            self.model,
            token=self.config.access_token,
            cache_dir=VOLUME_PATH_MODEL,
        )


@dataclass
class HuggingFaceDatasetParams:
    repo_id: str
    access_token: Optional[str] = None
    # TODO (andreyvelich): Discuss where we should specify dataset preprocess parameters.
    split: Optional[str] = None

    def __post_init__(self):
        # Custom checks or validations can be added here
        if self.repo_id == "" or self.repo_id is None:
            raise ValueError("repo_id is None")


class HuggingFaceDataset(datasetProvider):
    def load_config(self, serialised_args):
        self.config = HuggingFaceDatasetParams(**json.loads(serialised_args))

    def download_dataset(self):
        logger.info("Downloading dataset")
        logger.info("-" * 40)
        import huggingface_hub
        from datasets import load_dataset

        if self.config.access_token:
            huggingface_hub.login(self.config.access_token)

        # Load dataset and save to disk.
        dataset = load_dataset(self.config.repo_id, split=self.config.split)
        dataset.save_to_disk(VOLUME_PATH_DATASET)



# Copyright 2021 kubeflow.org.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import logging
import pytest
from typing import Optional

from kubernetes.client import V1PodTemplateSpec
from kubernetes.client import V1ObjectMeta
from kubernetes.client import V1PodSpec
from kubernetes.client import V1Container
from kubernetes.client import V1ResourceRequirements

from kubeflow.training import TrainingClient
from kubeflow.training import KubeflowOrgV1ReplicaSpec
from kubeflow.training import KubeflowOrgV1XGBoostJob
from kubeflow.training import KubeflowOrgV1XGBoostJobSpec
from kubeflow.training import KubeflowOrgV1RunPolicy
from kubeflow.training import KubeflowOrgV1SchedulingPolicy
from kubeflow.training.constants import constants

import test.e2e.utils as utils
from test.e2e.constants import TEST_GANG_SCHEDULER_NAME_ENV_KEY
from test.e2e.constants import GANG_SCHEDULERS, NONE_GANG_SCHEDULERS

logging.basicConfig(format="%(message)s")
logging.getLogger("kubeflow.training.api.training_client").setLevel(logging.DEBUG)

TRAINING_CLIENT = TrainingClient(job_kind=constants.XGBOOSTJOB_KIND)
JOB_NAME = "xgboostjob-iris-ci-test"
CONTAINER_NAME = "xgboost"
GANG_SCHEDULER_NAME = os.getenv(TEST_GANG_SCHEDULER_NAME_ENV_KEY, "")


@pytest.mark.skipif(
    GANG_SCHEDULER_NAME in NONE_GANG_SCHEDULERS,
    reason="For gang-scheduling",
)
def test_sdk_e2e_with_gang_scheduling(job_namespace):
    container = generate_container()

    master = KubeflowOrgV1ReplicaSpec(
        replicas=1,
        restart_policy="OnFailure",
        template=V1PodTemplateSpec(
            metadata=V1ObjectMeta(
                annotations={constants.ISTIO_SIDECAR_INJECTION: "false"}
            ),
            spec=V1PodSpec(
                containers=[container],
                scheduler_name=utils.get_pod_spec_scheduler_name(GANG_SCHEDULER_NAME),
            ),
        ),
    )

    worker = KubeflowOrgV1ReplicaSpec(
        replicas=1,
        restart_policy="OnFailure",
        template=V1PodTemplateSpec(
            metadata=V1ObjectMeta(
                annotations={constants.ISTIO_SIDECAR_INJECTION: "false"}
            ),
            spec=V1PodSpec(
                containers=[container],
                scheduler_name=utils.get_pod_spec_scheduler_name(GANG_SCHEDULER_NAME),
            ),
        ),
    )

    unschedulable_xgboostjob = generate_xgboostjob(
        job_namespace, master, worker, KubeflowOrgV1SchedulingPolicy(min_available=10)
    )
    schedulable_xgboostjob = generate_xgboostjob(
        job_namespace, master, worker, KubeflowOrgV1SchedulingPolicy(min_available=2)
    )

    TRAINING_CLIENT.create_job(job=unschedulable_xgboostjob, namespace=job_namespace)
    logging.info(f"List of created {TRAINING_CLIENT.job_kind}s")
    logging.info(TRAINING_CLIENT.list_jobs(job_namespace))

    try:
        utils.verify_unschedulable_job_e2e(TRAINING_CLIENT, JOB_NAME, job_namespace)
    except Exception as e:
        utils.print_job_results(TRAINING_CLIENT, JOB_NAME, job_namespace)
        TRAINING_CLIENT.delete_job(JOB_NAME, job_namespace)
        raise Exception(f"XGBoostJob E2E fails. Exception: {e}")

    TRAINING_CLIENT.update_job(schedulable_xgboostjob, JOB_NAME, job_namespace)
    logging.info(f"List of updated {TRAINING_CLIENT.job_kind}s")
    logging.info(TRAINING_CLIENT.list_jobs(job_namespace))

    try:
        utils.verify_job_e2e(TRAINING_CLIENT, JOB_NAME, job_namespace, wait_timeout=900)
    except Exception as e:
        utils.print_job_results(TRAINING_CLIENT, JOB_NAME, job_namespace)
        TRAINING_CLIENT.delete_job(JOB_NAME, job_namespace)
        raise Exception(f"XGBoostJob E2E fails. Exception: {e}")

    utils.print_job_results(TRAINING_CLIENT, JOB_NAME, job_namespace)
    TRAINING_CLIENT.delete_job(JOB_NAME, job_namespace)


@pytest.mark.skipif(
    GANG_SCHEDULER_NAME in GANG_SCHEDULERS,
    reason="For plain scheduling",
)
def test_sdk_e2e(job_namespace):
    container = generate_container()

    master = KubeflowOrgV1ReplicaSpec(
        replicas=1,
        restart_policy="OnFailure",
        template=V1PodTemplateSpec(
            metadata=V1ObjectMeta(
                annotations={constants.ISTIO_SIDECAR_INJECTION: "false"}
            ),
            spec=V1PodSpec(containers=[container]),
        ),
    )

    worker = KubeflowOrgV1ReplicaSpec(
        replicas=1,
        restart_policy="OnFailure",
        template=V1PodTemplateSpec(
            metadata=V1ObjectMeta(
                annotations={constants.ISTIO_SIDECAR_INJECTION: "false"}
            ),
            spec=V1PodSpec(containers=[container]),
        ),
    )

    xgboostjob = generate_xgboostjob(job_namespace, master, worker)

    TRAINING_CLIENT.create_job(job=xgboostjob, namespace=job_namespace)
    logging.info(f"List of created {TRAINING_CLIENT.job_kind}s")
    logging.info(TRAINING_CLIENT.list_jobs(job_namespace))

    try:
        utils.verify_job_e2e(TRAINING_CLIENT, JOB_NAME, job_namespace, wait_timeout=900)
    except Exception as e:
        utils.print_job_results(TRAINING_CLIENT, JOB_NAME, job_namespace)
        TRAINING_CLIENT.delete_job(JOB_NAME, job_namespace)
        raise Exception(f"XGBoostJob E2E fails. Exception: {e}")

    utils.print_job_results(TRAINING_CLIENT, JOB_NAME, job_namespace)
    TRAINING_CLIENT.delete_job(JOB_NAME, job_namespace)


def generate_xgboostjob(
    job_namespace: str,
    master: KubeflowOrgV1ReplicaSpec,
    worker: KubeflowOrgV1ReplicaSpec,
    scheduling_policy: Optional[KubeflowOrgV1SchedulingPolicy] = None,
) -> KubeflowOrgV1XGBoostJob:
    return KubeflowOrgV1XGBoostJob(
        api_version=constants.API_VERSION,
        kind=constants.XGBOOSTJOB_KIND,
        metadata=V1ObjectMeta(name=JOB_NAME, namespace=job_namespace),
        spec=KubeflowOrgV1XGBoostJobSpec(
            run_policy=KubeflowOrgV1RunPolicy(
                clean_pod_policy="None",
                scheduling_policy=scheduling_policy,
            ),
            xgb_replica_specs={"Master": master, "Worker": worker},
        ),
    )


def generate_container() -> V1Container:
    return V1Container(
        name=CONTAINER_NAME,
        image="docker.io/kubeflow/xgboost-dist-iris:latest",
        args=[
            "--job_type=Train",
            "--xgboost_parameter=objective:multi:softprob,num_class:3",
            "--n_estimators=10",
            "--learning_rate=0.1",
            "--model_path=/tmp/xgboost-model",
            "--model_storage_type=local",
        ],
        resources=V1ResourceRequirements(limits={"memory": "2Gi", "cpu": "0.8"}),
    )



from abc import ABC, abstractmethod


class modelProvider(ABC):
    @abstractmethod
    def load_config(self):
        pass

    @abstractmethod
    def download_model_and_tokenizer(self):
        pass



import logging
import time

from kubeflow.training import TrainingClient
from kubeflow.training.constants import constants
from test.e2e.constants import TEST_GANG_SCHEDULER_NAME_SCHEDULER_PLUGINS
from test.e2e.constants import DEFAULT_SCHEDULER_PLUGINS_NAME
from test.e2e.constants import TEST_GANG_SCHEDULER_NAME_VOLCANO

logging.basicConfig(format="%(message)s")
logging.getLogger().setLevel(logging.INFO)


def verify_unschedulable_job_e2e(client: TrainingClient, name: str, namespace: str):
    """Verify unschedulable Training Job e2e test."""
    logging.info(f"\n\n\n{client.job_kind} is creating")
    job = client.wait_for_job_conditions(
        name, namespace, expected_conditions={constants.JOB_CONDITION_CREATED}
    )

    logging.info("Checking 3 times that pods are not scheduled")
    for num in range(3):
        logging.info(f"Number of attempts: {int(num)+1}/3")

        # Job should have correct conditions
        if not client.is_job_created(job=job) or client.is_job_running(job=job):
            raise Exception(
                f"{client.job_kind} should be in Created condition. "
                f"{client.job_kind} should not be in Running condition."
            )

        logging.info("Sleeping 5 seconds...")
        time.sleep(5)


def verify_job_e2e(
    client: TrainingClient,
    name: str,
    namespace: str,
    wait_timeout: int = 600,
):
    """Verify Training Job e2e test."""

    # Wait until Job is Succeeded.
    logging.info(f"\n\n\n{client.job_kind} is running")
    job = client.wait_for_job_conditions(name, namespace, wait_timeout=wait_timeout)

    # Job should have Created, Running, and Succeeded conditions.
    conditions = client.get_job_conditions(job=job)
    # If Job is complete fast, it has 2 conditions: Created and Succeeded.
    if len(conditions) < 2:
        raise Exception(f"{client.job_kind} conditions are invalid: {conditions}")

    # Job should have correct conditions.
    if (
        not client.is_job_created(job=job)
        or not client.is_job_succeeded(job=job)
        or client.is_job_running(job=job)
        or client.is_job_restarting(job=job)
        or client.is_job_failed(job=job)
    ):
        raise Exception(
            f"{client.job_kind} should be in Succeeded and Created conditions. "
            f"{client.job_kind} should not be in Running, Restarting, or Failed conditions."
        )


def get_pod_spec_scheduler_name(gang_scheduler_name: str) -> str:
    if gang_scheduler_name == TEST_GANG_SCHEDULER_NAME_SCHEDULER_PLUGINS:
        return DEFAULT_SCHEDULER_PLUGINS_NAME
    elif gang_scheduler_name == TEST_GANG_SCHEDULER_NAME_VOLCANO:
        return TEST_GANG_SCHEDULER_NAME_VOLCANO

    return ""


def print_job_results(client: TrainingClient, name: str, namespace: str):
    # Print Job.
    logging.info(f"\n\n\n{client.job_kind} info")
    logging.info(client.get_job(name, namespace))

    # Print Job pod names.
    logging.info(f"\n\n\n{client.job_kind} pod names")
    logging.info(client.get_job_pod_names(name, namespace))

    # Print Job logs.
    logging.info(f"\n\n\n{client.job_kind} logs")
    client.get_job_logs(name, namespace)



# coding: utf-8

"""
    Kubeflow Training SDK

    Python SDK for Kubeflow Training  # noqa: E501

    The version of the OpenAPI document: v1.9.2
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from kubeflow.training.configuration import Configuration


class KubeflowOrgV1RunPolicy(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'active_deadline_seconds': 'int',
        'backoff_limit': 'int',
        'clean_pod_policy': 'str',
        'managed_by': 'str',
        'scheduling_policy': 'KubeflowOrgV1SchedulingPolicy',
        'suspend': 'bool',
        'ttl_seconds_after_finished': 'int'
    }

    attribute_map = {
        'active_deadline_seconds': 'activeDeadlineSeconds',
        'backoff_limit': 'backoffLimit',
        'clean_pod_policy': 'cleanPodPolicy',
        'managed_by': 'managedBy',
        'scheduling_policy': 'schedulingPolicy',
        'suspend': 'suspend',
        'ttl_seconds_after_finished': 'ttlSecondsAfterFinished'
    }

    def __init__(self, active_deadline_seconds=None, backoff_limit=None, clean_pod_policy=None, managed_by=None, scheduling_policy=None, suspend=None, ttl_seconds_after_finished=None, local_vars_configuration=None):  # noqa: E501
        """KubeflowOrgV1RunPolicy - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._active_deadline_seconds = None
        self._backoff_limit = None
        self._clean_pod_policy = None
        self._managed_by = None
        self._scheduling_policy = None
        self._suspend = None
        self._ttl_seconds_after_finished = None
        self.discriminator = None

        if active_deadline_seconds is not None:
            self.active_deadline_seconds = active_deadline_seconds
        if backoff_limit is not None:
            self.backoff_limit = backoff_limit
        if clean_pod_policy is not None:
            self.clean_pod_policy = clean_pod_policy
        if managed_by is not None:
            self.managed_by = managed_by
        if scheduling_policy is not None:
            self.scheduling_policy = scheduling_policy
        if suspend is not None:
            self.suspend = suspend
        if ttl_seconds_after_finished is not None:
            self.ttl_seconds_after_finished = ttl_seconds_after_finished

    @property
    def active_deadline_seconds(self):
        """Gets the active_deadline_seconds of this KubeflowOrgV1RunPolicy.  # noqa: E501

        Specifies the duration in seconds relative to the startTime that the job may be active before the system tries to terminate it; value must be positive integer.  # noqa: E501

        :return: The active_deadline_seconds of this KubeflowOrgV1RunPolicy.  # noqa: E501
        :rtype: int
        """
        return self._active_deadline_seconds

    @active_deadline_seconds.setter
    def active_deadline_seconds(self, active_deadline_seconds):
        """Sets the active_deadline_seconds of this KubeflowOrgV1RunPolicy.

        Specifies the duration in seconds relative to the startTime that the job may be active before the system tries to terminate it; value must be positive integer.  # noqa: E501

        :param active_deadline_seconds: The active_deadline_seconds of this KubeflowOrgV1RunPolicy.  # noqa: E501
        :type: int
        """

        self._active_deadline_seconds = active_deadline_seconds

    @property
    def backoff_limit(self):
        """Gets the backoff_limit of this KubeflowOrgV1RunPolicy.  # noqa: E501

        Optional number of retries before marking this job failed.  # noqa: E501

        :return: The backoff_limit of this KubeflowOrgV1RunPolicy.  # noqa: E501
        :rtype: int
        """
        return self._backoff_limit

    @backoff_limit.setter
    def backoff_limit(self, backoff_limit):
        """Sets the backoff_limit of this KubeflowOrgV1RunPolicy.

        Optional number of retries before marking this job failed.  # noqa: E501

        :param backoff_limit: The backoff_limit of this KubeflowOrgV1RunPolicy.  # noqa: E501
        :type: int
        """

        self._backoff_limit = backoff_limit

    @property
    def clean_pod_policy(self):
        """Gets the clean_pod_policy of this KubeflowOrgV1RunPolicy.  # noqa: E501

        CleanPodPolicy defines the policy to kill pods after the job completes. Default to None.  # noqa: E501

        :return: The clean_pod_policy of this KubeflowOrgV1RunPolicy.  # noqa: E501
        :rtype: str
        """
        return self._clean_pod_policy

    @clean_pod_policy.setter
    def clean_pod_policy(self, clean_pod_policy):
        """Sets the clean_pod_policy of this KubeflowOrgV1RunPolicy.

        CleanPodPolicy defines the policy to kill pods after the job completes. Default to None.  # noqa: E501

        :param clean_pod_policy: The clean_pod_policy of this KubeflowOrgV1RunPolicy.  # noqa: E501
        :type: str
        """

        self._clean_pod_policy = clean_pod_policy

    @property
    def managed_by(self):
        """Gets the managed_by of this KubeflowOrgV1RunPolicy.  # noqa: E501

        ManagedBy is used to indicate the controller or entity that manages a job. The value must be either an empty, 'kubeflow.org/training-operator' or 'kueue.x-k8s.io/multikueue'. The training-operator reconciles a job which doesn't have this field at all or the field value is the reserved string 'kubeflow.org/training-operator', but delegates reconciling the job with 'kueue.x-k8s.io/multikueue' to the Kueue. The field is immutable.  # noqa: E501

        :return: The managed_by of this KubeflowOrgV1RunPolicy.  # noqa: E501
        :rtype: str
        """
        return self._managed_by

    @managed_by.setter
    def managed_by(self, managed_by):
        """Sets the managed_by of this KubeflowOrgV1RunPolicy.

        ManagedBy is used to indicate the controller or entity that manages a job. The value must be either an empty, 'kubeflow.org/training-operator' or 'kueue.x-k8s.io/multikueue'. The training-operator reconciles a job which doesn't have this field at all or the field value is the reserved string 'kubeflow.org/training-operator', but delegates reconciling the job with 'kueue.x-k8s.io/multikueue' to the Kueue. The field is immutable.  # noqa: E501

        :param managed_by: The managed_by of this KubeflowOrgV1RunPolicy.  # noqa: E501
        :type: str
        """

        self._managed_by = managed_by

    @property
    def scheduling_policy(self):
        """Gets the scheduling_policy of this KubeflowOrgV1RunPolicy.  # noqa: E501


        :return: The scheduling_policy of this KubeflowOrgV1RunPolicy.  # noqa: E501
        :rtype: KubeflowOrgV1SchedulingPolicy
        """
        return self._scheduling_policy

    @scheduling_policy.setter
    def scheduling_policy(self, scheduling_policy):
        """Sets the scheduling_policy of this KubeflowOrgV1RunPolicy.


        :param scheduling_policy: The scheduling_policy of this KubeflowOrgV1RunPolicy.  # noqa: E501
        :type: KubeflowOrgV1SchedulingPolicy
        """

        self._scheduling_policy = scheduling_policy

    @property
    def suspend(self):
        """Gets the suspend of this KubeflowOrgV1RunPolicy.  # noqa: E501

        suspend specifies whether the Job controller should create Pods or not. If a Job is created with suspend set to true, no Pods are created by the Job controller. If a Job is suspended after creation (i.e. the flag goes from false to true), the Job controller will delete all active Pods and PodGroups associated with this Job. Users must design their workload to gracefully handle this. Suspending a Job will reset the StartTime field of the Job.  Defaults to false.  # noqa: E501

        :return: The suspend of this KubeflowOrgV1RunPolicy.  # noqa: E501
        :rtype: bool
        """
        return self._suspend

    @suspend.setter
    def suspend(self, suspend):
        """Sets the suspend of this KubeflowOrgV1RunPolicy.

        suspend specifies whether the Job controller should create Pods or not. If a Job is created with suspend set to true, no Pods are created by the Job controller. If a Job is suspended after creation (i.e. the flag goes from false to true), the Job controller will delete all active Pods and PodGroups associated with this Job. Users must design their workload to gracefully handle this. Suspending a Job will reset the StartTime field of the Job.  Defaults to false.  # noqa: E501

        :param suspend: The suspend of this KubeflowOrgV1RunPolicy.  # noqa: E501
        :type: bool
        """

        self._suspend = suspend

    @property
    def ttl_seconds_after_finished(self):
        """Gets the ttl_seconds_after_finished of this KubeflowOrgV1RunPolicy.  # noqa: E501

        TTLSecondsAfterFinished is the TTL to clean up jobs. It may take extra ReconcilePeriod seconds for the cleanup, since reconcile gets called periodically. Default to infinite.  # noqa: E501

        :return: The ttl_seconds_after_finished of this KubeflowOrgV1RunPolicy.  # noqa: E501
        :rtype: int
        """
        return self._ttl_seconds_after_finished

    @ttl_seconds_after_finished.setter
    def ttl_seconds_after_finished(self, ttl_seconds_after_finished):
        """Sets the ttl_seconds_after_finished of this KubeflowOrgV1RunPolicy.

        TTLSecondsAfterFinished is the TTL to clean up jobs. It may take extra ReconcilePeriod seconds for the cleanup, since reconcile gets called periodically. Default to infinite.  # noqa: E501

        :param ttl_seconds_after_finished: The ttl_seconds_after_finished of this KubeflowOrgV1RunPolicy.  # noqa: E501
        :type: int
        """

        self._ttl_seconds_after_finished = ttl_seconds_after_finished

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, KubeflowOrgV1RunPolicy):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, KubeflowOrgV1RunPolicy):
            return True

        return self.to_dict() != other.to_dict()



# Copyright 2021 kubeflow.org.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Union

from kubeflow.storage_initializer.constants import INIT_CONTAINER_MOUNT_PATH
from kubeflow.training import models

# How long to wait in seconds for requests to the Kubernetes API Server.
DEFAULT_TIMEOUT = 120

# The default PIP index URL to download Python packages.
DEFAULT_PIP_INDEX_URL = "https://pypi.org/simple"

# The default namespace in case namespace not define explicitly
DEFAULT_NAMESPACE = "default"

# Annotation to disable Istio sidecar.
ISTIO_SIDECAR_INJECTION = "sidecar.istio.io/inject"

# Common constants.
GROUP = "kubeflow.org"
VERSION = "v1"
API_VERSION = f"{GROUP}/{VERSION}"

# Kind for pod.
POD_KIND = "Pod"

# Pending status for pod phase.
POD_PHASE_PENDING = "Pending"


# Training Job conditions.
JOB_CONDITION_CREATED = "Created"
JOB_CONDITION_RUNNING = "Running"
JOB_CONDITION_RESTARTING = "Restarting"
JOB_CONDITION_SUCCEEDED = "Succeeded"
JOB_CONDITION_FAILED = "Failed"
JOB_CONDITIONS = {
    JOB_CONDITION_CREATED,
    JOB_CONDITION_RUNNING,
    JOB_CONDITION_RESTARTING,
    JOB_CONDITION_SUCCEEDED,
    JOB_CONDITION_FAILED,
}
# True means that Training Job is in this condition.
CONDITION_STATUS_TRUE = "True"

# Job Label Names
JOB_NAME_LABEL = "training.kubeflow.org/job-name"
JOB_ROLE_LABEL = "training.kubeflow.org/job-role"
JOB_ROLE_MASTER = "master"
REPLICA_TYPE_LABEL = "training.kubeflow.org/replica-type"
REPLICA_INDEX_LABEL = "training.kubeflow.org/replica-index"

# Various replica types.
REPLICA_TYPE_CHIEF = "Chief"
REPLICA_TYPE_PS = "PS"
REPLICA_TYPE_MASTER = "Master"
REPLICA_TYPE_WORKER = "Worker"
REPLICA_TYPE_SCHEDULER = "Scheduler"
REPLICA_TYPE_SERVER = "Server"
REPLICA_TYPE_LAUNCHER = "Launcher"

# Constants for Train API.
STORAGE_INITIALIZER = "storage-initializer"
# The default value for dataset and model storage PVC.
PVC_DEFAULT_SIZE = "10Gi"
# The default value for PVC access modes.
PVC_DEFAULT_ACCESS_MODES = ["ReadWriteOnce", "ReadOnlyMany"]


# TODO (andreyvelich): We should add image tag for Storage Initializer and Trainer.
STORAGE_INITIALIZER_IMAGE = os.getenv(
    "STORAGE_INITIALIZER_IMAGE", "ghcr.io/kubeflow/training-v1/storage-initializer"
)

STORAGE_INITIALIZER_VOLUME_MOUNT = models.V1VolumeMount(
    name=STORAGE_INITIALIZER,
    mount_path=INIT_CONTAINER_MOUNT_PATH,
)

TRAINER_TRANSFORMER_IMAGE = os.getenv(
    "TRAINER_TRANSFORMER_IMAGE", "ghcr.io/kubeflow/training-v1/trainer-huggingface"
)

# TFJob constants.
TFJOB_KIND = "TFJob"
TFJOB_MODEL = "KubeflowOrgV1TFJob"
TFJOB_PLURAL = "tfjobs"
TFJOB_CONTAINER = "tensorflow"
TFJOB_REPLICA_TYPES = (
    REPLICA_TYPE_PS.lower(),
    REPLICA_TYPE_CHIEF.lower(),
    REPLICA_TYPE_WORKER.lower(),
)

TFJOB_BASE_IMAGE = "docker.io/tensorflow/tensorflow:2.9.1"
TFJOB_BASE_IMAGE_GPU = "docker.io/tensorflow/tensorflow:2.9.1-gpu"

# PyTorchJob constants
PYTORCHJOB_KIND = "PyTorchJob"
PYTORCHJOB_MODEL = "KubeflowOrgV1PyTorchJob"
PYTORCHJOB_PLURAL = "pytorchjobs"
PYTORCHJOB_CONTAINER = "pytorch"
PYTORCHJOB_REPLICA_TYPES = (REPLICA_TYPE_MASTER.lower(), REPLICA_TYPE_WORKER.lower())
PYTORCHJOB_BASE_IMAGE = "docker.io/pytorch/pytorch:2.1.2-cuda11.8-cudnn8-runtime"

ENTRYPOINT_TORCH = "torchrun"
ENTRYPOINT_PYTHON = "python -u"
DEFAULT_COMMAND = ["bash", "-c"]

# XGBoostJob constants
XGBOOSTJOB_KIND = "XGBoostJob"
XGBOOSTJOB_MODEL = "KubeflowOrgV1XGBoostJob"
XGBOOSTJOB_PLURAL = "xgboostjobs"
XGBOOSTJOB_CONTAINER = "xgboost"
XGBOOSTJOB_REPLICA_TYPES = (REPLICA_TYPE_MASTER.lower(), REPLICA_TYPE_WORKER.lower())

# MPIJob constants
MPIJOB_KIND = "MPIJob"
MPIJOB_MODEL = "KubeflowOrgV1MPIJob"
MPIJOB_PLURAL = "mpijobs"
MPIJOB_CONTAINER = "mpi"
MPIJOB_REPLICA_TYPES = (REPLICA_TYPE_LAUNCHER.lower(), REPLICA_TYPE_WORKER.lower())

# PaddleJob constants
PADDLEJOB_KIND = "PaddleJob"
PADDLEJOB_MODEL = "KubeflowOrgV1PaddleJob"
PADDLEJOB_PLURAL = "paddlejobs"
PADDLEJOB_CONTAINER = "paddle"
PADDLEJOB_REPLICA_TYPES = (REPLICA_TYPE_MASTER.lower(), REPLICA_TYPE_WORKER.lower())

PADDLEJOB_BASE_IMAGE = (
    "docker.io/paddlepaddle/paddle:2.4.0rc0-gpu-cuda11.2-cudnn8.1-trt8.0"
)

# JAXJob constants
JAXJOB_KIND = "JAXJob"
JAXJOB_MODEL = "KubeflowOrgV1JAXJob"
JAXJOB_PLURAL = "jaxjobs"
JAXJOB_CONTAINER = "jax"
JAXJOB_REPLICA_TYPES = REPLICA_TYPE_WORKER.lower()
JAXJOB_BASE_IMAGE = "ghcr.io/kubeflow/training-v1/jaxjob-dist-spmd-mnist:latest"

# Dictionary to get plural, model, and container for each Job kind.
JOB_PARAMETERS = {
    TFJOB_KIND: {
        "model": TFJOB_MODEL,
        "plural": TFJOB_PLURAL,
        "container": TFJOB_CONTAINER,
        "base_image": TFJOB_BASE_IMAGE,
    },
    PYTORCHJOB_KIND: {
        "model": PYTORCHJOB_MODEL,
        "plural": PYTORCHJOB_PLURAL,
        "container": PYTORCHJOB_CONTAINER,
        "base_image": PYTORCHJOB_BASE_IMAGE,
    },
    XGBOOSTJOB_KIND: {
        "model": XGBOOSTJOB_MODEL,
        "plural": XGBOOSTJOB_PLURAL,
        "container": XGBOOSTJOB_CONTAINER,
        "base_image": "TODO",
    },
    MPIJOB_KIND: {
        "model": MPIJOB_MODEL,
        "plural": MPIJOB_PLURAL,
        "container": MPIJOB_CONTAINER,
        "base_image": "TODO",
    },
    PADDLEJOB_KIND: {
        "model": PADDLEJOB_MODEL,
        "plural": PADDLEJOB_PLURAL,
        "container": PADDLEJOB_CONTAINER,
        "base_image": PADDLEJOB_BASE_IMAGE,
    },
    JAXJOB_KIND: {
        "model": JAXJOB_MODEL,
        "plural": JAXJOB_PLURAL,
        "container": JAXJOB_CONTAINER,
        "base_image": "JAXJOB_BASE_IMAGE",
    },
}

# Tuple of all Job models.
JOB_MODELS = tuple([d["model"] for d in list(JOB_PARAMETERS.values())])

# Union type of all Job models.
JOB_MODELS_TYPE = Union[
    models.KubeflowOrgV1TFJob,
    models.KubeflowOrgV1PyTorchJob,
    models.KubeflowOrgV1XGBoostJob,
    models.KubeflowOrgV1MPIJob,
    models.KubeflowOrgV1PaddleJob,
    models.KubeflowOrgV1JAXJob,
]



# coding: utf-8

"""
    Kubeflow Training SDK

    Python SDK for Kubeflow Training  # noqa: E501

    The version of the OpenAPI document: v1.9.2
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from kubeflow.training.configuration import Configuration


class KubeflowOrgV1MPIJobSpec(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'clean_pod_policy': 'str',
        'main_container': 'str',
        'mpi_replica_specs': 'dict(str, KubeflowOrgV1ReplicaSpec)',
        'run_policy': 'KubeflowOrgV1RunPolicy',
        'slots_per_worker': 'int'
    }

    attribute_map = {
        'clean_pod_policy': 'cleanPodPolicy',
        'main_container': 'mainContainer',
        'mpi_replica_specs': 'mpiReplicaSpecs',
        'run_policy': 'runPolicy',
        'slots_per_worker': 'slotsPerWorker'
    }

    def __init__(self, clean_pod_policy=None, main_container=None, mpi_replica_specs=None, run_policy=None, slots_per_worker=None, local_vars_configuration=None):  # noqa: E501
        """KubeflowOrgV1MPIJobSpec - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._clean_pod_policy = None
        self._main_container = None
        self._mpi_replica_specs = None
        self._run_policy = None
        self._slots_per_worker = None
        self.discriminator = None

        if clean_pod_policy is not None:
            self.clean_pod_policy = clean_pod_policy
        if main_container is not None:
            self.main_container = main_container
        self.mpi_replica_specs = mpi_replica_specs
        if run_policy is not None:
            self.run_policy = run_policy
        if slots_per_worker is not None:
            self.slots_per_worker = slots_per_worker

    @property
    def clean_pod_policy(self):
        """Gets the clean_pod_policy of this KubeflowOrgV1MPIJobSpec.  # noqa: E501

        CleanPodPolicy defines the policy that whether to kill pods after the job completes. Defaults to None.  # noqa: E501

        :return: The clean_pod_policy of this KubeflowOrgV1MPIJobSpec.  # noqa: E501
        :rtype: str
        """
        return self._clean_pod_policy

    @clean_pod_policy.setter
    def clean_pod_policy(self, clean_pod_policy):
        """Sets the clean_pod_policy of this KubeflowOrgV1MPIJobSpec.

        CleanPodPolicy defines the policy that whether to kill pods after the job completes. Defaults to None.  # noqa: E501

        :param clean_pod_policy: The clean_pod_policy of this KubeflowOrgV1MPIJobSpec.  # noqa: E501
        :type: str
        """

        self._clean_pod_policy = clean_pod_policy

    @property
    def main_container(self):
        """Gets the main_container of this KubeflowOrgV1MPIJobSpec.  # noqa: E501

        MainContainer specifies name of the main container which executes the MPI code.  # noqa: E501

        :return: The main_container of this KubeflowOrgV1MPIJobSpec.  # noqa: E501
        :rtype: str
        """
        return self._main_container

    @main_container.setter
    def main_container(self, main_container):
        """Sets the main_container of this KubeflowOrgV1MPIJobSpec.

        MainContainer specifies name of the main container which executes the MPI code.  # noqa: E501

        :param main_container: The main_container of this KubeflowOrgV1MPIJobSpec.  # noqa: E501
        :type: str
        """

        self._main_container = main_container

    @property
    def mpi_replica_specs(self):
        """Gets the mpi_replica_specs of this KubeflowOrgV1MPIJobSpec.  # noqa: E501

        `MPIReplicaSpecs` contains maps from `MPIReplicaType` to `ReplicaSpec` that specify the MPI replicas to run.  # noqa: E501

        :return: The mpi_replica_specs of this KubeflowOrgV1MPIJobSpec.  # noqa: E501
        :rtype: dict(str, KubeflowOrgV1ReplicaSpec)
        """
        return self._mpi_replica_specs

    @mpi_replica_specs.setter
    def mpi_replica_specs(self, mpi_replica_specs):
        """Sets the mpi_replica_specs of this KubeflowOrgV1MPIJobSpec.

        `MPIReplicaSpecs` contains maps from `MPIReplicaType` to `ReplicaSpec` that specify the MPI replicas to run.  # noqa: E501

        :param mpi_replica_specs: The mpi_replica_specs of this KubeflowOrgV1MPIJobSpec.  # noqa: E501
        :type: dict(str, KubeflowOrgV1ReplicaSpec)
        """
        if self.local_vars_configuration.client_side_validation and mpi_replica_specs is None:  # noqa: E501
            raise ValueError("Invalid value for `mpi_replica_specs`, must not be `None`")  # noqa: E501

        self._mpi_replica_specs = mpi_replica_specs

    @property
    def run_policy(self):
        """Gets the run_policy of this KubeflowOrgV1MPIJobSpec.  # noqa: E501


        :return: The run_policy of this KubeflowOrgV1MPIJobSpec.  # noqa: E501
        :rtype: KubeflowOrgV1RunPolicy
        """
        return self._run_policy

    @run_policy.setter
    def run_policy(self, run_policy):
        """Sets the run_policy of this KubeflowOrgV1MPIJobSpec.


        :param run_policy: The run_policy of this KubeflowOrgV1MPIJobSpec.  # noqa: E501
        :type: KubeflowOrgV1RunPolicy
        """

        self._run_policy = run_policy

    @property
    def slots_per_worker(self):
        """Gets the slots_per_worker of this KubeflowOrgV1MPIJobSpec.  # noqa: E501

        Specifies the number of slots per worker used in hostfile. Defaults to 1.  # noqa: E501

        :return: The slots_per_worker of this KubeflowOrgV1MPIJobSpec.  # noqa: E501
        :rtype: int
        """
        return self._slots_per_worker

    @slots_per_worker.setter
    def slots_per_worker(self, slots_per_worker):
        """Sets the slots_per_worker of this KubeflowOrgV1MPIJobSpec.

        Specifies the number of slots per worker used in hostfile. Defaults to 1.  # noqa: E501

        :param slots_per_worker: The slots_per_worker of this KubeflowOrgV1MPIJobSpec.  # noqa: E501
        :type: int
        """

        self._slots_per_worker = slots_per_worker

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, KubeflowOrgV1MPIJobSpec):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, KubeflowOrgV1MPIJobSpec):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Kubeflow Training SDK

    Python SDK for Kubeflow Training  # noqa: E501

    The version of the OpenAPI document: v1.9.2
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from kubeflow.training.configuration import Configuration


class KubeflowOrgV1PaddleElasticPolicy(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'max_replicas': 'int',
        'max_restarts': 'int',
        'metrics': 'list[K8sIoApiAutoscalingV2MetricSpec]',
        'min_replicas': 'int'
    }

    attribute_map = {
        'max_replicas': 'maxReplicas',
        'max_restarts': 'maxRestarts',
        'metrics': 'metrics',
        'min_replicas': 'minReplicas'
    }

    def __init__(self, max_replicas=None, max_restarts=None, metrics=None, min_replicas=None, local_vars_configuration=None):  # noqa: E501
        """KubeflowOrgV1PaddleElasticPolicy - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._max_replicas = None
        self._max_restarts = None
        self._metrics = None
        self._min_replicas = None
        self.discriminator = None

        if max_replicas is not None:
            self.max_replicas = max_replicas
        if max_restarts is not None:
            self.max_restarts = max_restarts
        if metrics is not None:
            self.metrics = metrics
        if min_replicas is not None:
            self.min_replicas = min_replicas

    @property
    def max_replicas(self):
        """Gets the max_replicas of this KubeflowOrgV1PaddleElasticPolicy.  # noqa: E501

        upper limit for the number of pods that can be set by the autoscaler; cannot be smaller than MinReplicas, defaults to null.  # noqa: E501

        :return: The max_replicas of this KubeflowOrgV1PaddleElasticPolicy.  # noqa: E501
        :rtype: int
        """
        return self._max_replicas

    @max_replicas.setter
    def max_replicas(self, max_replicas):
        """Sets the max_replicas of this KubeflowOrgV1PaddleElasticPolicy.

        upper limit for the number of pods that can be set by the autoscaler; cannot be smaller than MinReplicas, defaults to null.  # noqa: E501

        :param max_replicas: The max_replicas of this KubeflowOrgV1PaddleElasticPolicy.  # noqa: E501
        :type: int
        """

        self._max_replicas = max_replicas

    @property
    def max_restarts(self):
        """Gets the max_restarts of this KubeflowOrgV1PaddleElasticPolicy.  # noqa: E501

        MaxRestarts is the limit for restart times of pods in elastic mode.  # noqa: E501

        :return: The max_restarts of this KubeflowOrgV1PaddleElasticPolicy.  # noqa: E501
        :rtype: int
        """
        return self._max_restarts

    @max_restarts.setter
    def max_restarts(self, max_restarts):
        """Sets the max_restarts of this KubeflowOrgV1PaddleElasticPolicy.

        MaxRestarts is the limit for restart times of pods in elastic mode.  # noqa: E501

        :param max_restarts: The max_restarts of this KubeflowOrgV1PaddleElasticPolicy.  # noqa: E501
        :type: int
        """

        self._max_restarts = max_restarts

    @property
    def metrics(self):
        """Gets the metrics of this KubeflowOrgV1PaddleElasticPolicy.  # noqa: E501

        Metrics contains the specifications which are used to calculate the desired replica count (the maximum replica count across all metrics will be used).  The desired replica count is calculated with multiplying the ratio between the target value and the current value by the current number of pods. Ergo, metrics used must decrease as the pod count is increased, and vice-versa.  See the individual metric source types for more information about how each type of metric must respond. If not set, the HPA will not be created.  # noqa: E501

        :return: The metrics of this KubeflowOrgV1PaddleElasticPolicy.  # noqa: E501
        :rtype: list[K8sIoApiAutoscalingV2MetricSpec]
        """
        return self._metrics

    @metrics.setter
    def metrics(self, metrics):
        """Sets the metrics of this KubeflowOrgV1PaddleElasticPolicy.

        Metrics contains the specifications which are used to calculate the desired replica count (the maximum replica count across all metrics will be used).  The desired replica count is calculated with multiplying the ratio between the target value and the current value by the current number of pods. Ergo, metrics used must decrease as the pod count is increased, and vice-versa.  See the individual metric source types for more information about how each type of metric must respond. If not set, the HPA will not be created.  # noqa: E501

        :param metrics: The metrics of this KubeflowOrgV1PaddleElasticPolicy.  # noqa: E501
        :type: list[K8sIoApiAutoscalingV2MetricSpec]
        """

        self._metrics = metrics

    @property
    def min_replicas(self):
        """Gets the min_replicas of this KubeflowOrgV1PaddleElasticPolicy.  # noqa: E501

        minReplicas is the lower limit for the number of replicas to which the training job can scale down.  It defaults to null.  # noqa: E501

        :return: The min_replicas of this KubeflowOrgV1PaddleElasticPolicy.  # noqa: E501
        :rtype: int
        """
        return self._min_replicas

    @min_replicas.setter
    def min_replicas(self, min_replicas):
        """Sets the min_replicas of this KubeflowOrgV1PaddleElasticPolicy.

        minReplicas is the lower limit for the number of replicas to which the training job can scale down.  It defaults to null.  # noqa: E501

        :param min_replicas: The min_replicas of this KubeflowOrgV1PaddleElasticPolicy.  # noqa: E501
        :type: int
        """

        self._min_replicas = min_replicas

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, KubeflowOrgV1PaddleElasticPolicy):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, KubeflowOrgV1PaddleElasticPolicy):
            return True

        return self.to_dict() != other.to_dict()



import json
import os
from dataclasses import dataclass
from urllib.parse import urlparse

from .abstract_dataset_provider import datasetProvider
from .constants import VOLUME_PATH_DATASET


@dataclass
class S3DatasetParams:
    endpoint_url: str
    bucket_name: str
    file_key: str
    region_name: str = None
    access_key: str = None
    secret_key: str = None

    def is_valid_url(self, url):
        try:
            parsed_url = urlparse(url)
            print(parsed_url)
            return all([parsed_url.scheme, parsed_url.netloc])
        except ValueError:
            return False

    def __post_init__(self):
        # Custom checks or validations can be added here
        if (
            self.bucket_name is None
            or self.endpoint_url is None
            or self.file_key is None
        ):
            raise ValueError("bucket_name or endpoint_url or file_key is None")
        self.is_valid_url(self.endpoint_url)


class S3(datasetProvider):
    def load_config(self, serialised_args):
        self.config = S3DatasetParams(**json.loads(serialised_args))

    def download_dataset(self):
        import boto3

        # Create an S3 client for Nutanix Object Store/S3
        s3_client = boto3.Session(
            aws_access_key_id=self.config.access_key,
            aws_secret_access_key=self.config.secret_key,
            region_name=self.config.region_name,
        )
        s3_resource = s3_client.resource("s3", endpoint_url=self.config.endpoint_url)
        # Get the bucket object
        bucket = s3_resource.Bucket(self.config.bucket_name)

        # Filter objects with the specified prefix
        objects = bucket.objects.filter(Prefix=self.config.file_key)
        # Iterate over filtered objects
        for obj in objects:
            # Extract the object key (filename)
            obj_key = obj.key
            path_components = obj_key.split(os.path.sep)
            path_excluded_first_last_parts = os.path.sep.join(path_components[1:-1])

            # Create directories if they don't exist
            os.makedirs(
                os.path.join(VOLUME_PATH_DATASET, path_excluded_first_last_parts),
                exist_ok=True,
            )

            # Download the file
            file_path = os.path.sep.join(path_components[1:])
            bucket.download_file(obj_key, os.path.join(VOLUME_PATH_DATASET, file_path))
        print("Files downloaded")



import multiprocessing
import queue
from datetime import datetime, timedelta
from unittest.mock import Mock, patch

import pytest
from kubeflow.training import (
    KubeflowOrgV1JobCondition,
    KubeflowOrgV1JobStatus,
    KubeflowOrgV1PyTorchJob,
    KubeflowOrgV1PyTorchJobSpec,
    KubeflowOrgV1ReplicaSpec,
    KubeflowOrgV1RunPolicy,
    TrainingClient,
    constants,
)
from kubeflow.training.models import V1DeleteOptions
from kubernetes.client import (
    ApiClient,
    V1Container,
    V1EnvVar,
    V1ObjectMeta,
    V1PodSpec,
    V1PodTemplateSpec,
    V1Volume,
    V1VolumeMount,
)

TEST_NAME = "test"
TEST_IMAGE = "docker.io/test-training"

TIMEOUT = "timeout"
RUNTIME = "runtime"
MOCK_POD_OBJ = "mock_pod_obj"
NO_PODS = "no_pods"
DUMMY_POD_NAME = "Dummy V1PodList"
LIST_RESPONSE = [
    {"metadata": {"name": DUMMY_POD_NAME}},
]
SUCCESS = "success"
FAILED = "Failed"
CREATED = "Created"
RUNNING = "Running"
RESTARTING = "Restarting"
SUCCEEDED = "Succeeded"
INVALID = "invalid"

FAIL_LOGS = "fail_logs"
FAIL_EVENTS = "fail_events"
MULTI_PODS = "multi_pods"
PENDING_POD = "pending_pod"
NO_STATUS_POD = "no_status_pod"
QUEUE_TIMEOUT = "queue_timeout"
QUEUE_EMPTY = "queue_empty"
EVENT_CREATION_TIMESTAMP = datetime(2024, 1, 5, 22, 58, 20)


def conditional_error_handler(*args, **kwargs):
    if args[2] == TIMEOUT:
        raise multiprocessing.TimeoutError()
    elif args[2] == RUNTIME:
        raise RuntimeError()


def serialize_k8s_object(obj):
    api_client = ApiClient()
    return api_client.sanitize_for_serialization(obj)


def get_namespaced_custom_object_response(*args, **kwargs):
    if args[2] == TIMEOUT:
        raise multiprocessing.TimeoutError()
    elif args[2] == RUNTIME:
        raise RuntimeError()

    # Create a serialized Job
    serialized_job = serialize_k8s_object(generate_job_with_status(create_job()))

    # Mock the thread and set it's return value to the serialized Job
    mock_thread = Mock()
    mock_thread.get.return_value = serialized_job

    return mock_thread


def list_namespaced_custom_object_response(*args, **kwargs):
    if args[2] == TIMEOUT:
        raise multiprocessing.TimeoutError()
    elif args[2] == RUNTIME:
        raise RuntimeError()
    elif args[2] == "empty-namespace":
        mock_response = {"items": []}
    elif args[2] == "multi-jobs":
        mock_response = {
            "items": [
                serialize_k8s_object(generate_job_with_status(create_job())),
                serialize_k8s_object(generate_job_with_status(create_job())),
            ]
        }
    else:
        mock_response = {
            "items": [serialize_k8s_object(generate_job_with_status(create_job()))]
        }

    mock_thread = Mock()
    mock_thread.get.return_value = mock_response
    return mock_thread


def list_namespaced_pod_response(*args, **kwargs):
    class MockResponse:
        def get(self, timeout):
            """
            Simulates Kubernetes API response for listing namespaced pods,
            and pass timeout for verification

            :return:
                - If `args[0] == "timeout"`, raises `TimeoutError`.
                - If `args[0] == "runtime"`, raises `Exception`.
                - If `args[0] == "mock_pod_obj"`, returns a mock pod object
                  with `metadata.name = "Dummy V1PodList"`.
                - If `args[0] == "no_pods"`, returns an empty list of pods.
                - Otherwise, returns a default list of dicts representing pods,
                  with `timeout` included, for testing.
            """
            LIST_RESPONSE[0][TIMEOUT] = timeout
            if args[0] == TIMEOUT:
                raise multiprocessing.TimeoutError()
            if args[0] == RUNTIME:
                raise Exception()
            if args[0] == MOCK_POD_OBJ:
                pod_obj = Mock(metadata=Mock())
                pod_obj.metadata.name = DUMMY_POD_NAME
                return Mock(items=[pod_obj])
            if args[0] == NO_PODS:
                return Mock(items=[])
            return Mock(items=LIST_RESPONSE)

    return MockResponse()


def create_job(
    command=None,
    args=None,
    num_workers=2,
    labels=None,
    annotations=None,
    env_vars=None,
    volumes=None,
    volume_mounts=None,
):
    # Handle env_vars as either a dict or a list
    if env_vars:
        if isinstance(env_vars, dict):
            env_vars = [V1EnvVar(name=k, value=v) for k, v in env_vars.items()]
        elif isinstance(env_vars, list):
            env_vars = [
                v if isinstance(v, V1EnvVar) else V1EnvVar(**v) for v in env_vars
            ]

    container = V1Container(
        name=constants.PYTORCHJOB_CONTAINER,
        image=TEST_IMAGE,
        command=command,
        args=args,
        env=env_vars,
        volume_mounts=volume_mounts,
    )

    master = KubeflowOrgV1ReplicaSpec(
        replicas=1,
        template=V1PodTemplateSpec(
            metadata=V1ObjectMeta(
                annotations={constants.ISTIO_SIDECAR_INJECTION: "false"}
            ),
            spec=V1PodSpec(
                containers=[container],
                volumes=volumes,
            ),
        ),
    )

    pytorch_replica_specs = {"Master": master}

    # PyTorchJob always has 1 master and N-1 worker replicas.
    if num_workers > 1:
        pytorch_replica_specs["Worker"] = KubeflowOrgV1ReplicaSpec(
            replicas=num_workers - 1,
            template=V1PodTemplateSpec(
                metadata=V1ObjectMeta(
                    annotations={constants.ISTIO_SIDECAR_INJECTION: "false"}
                ),
                spec=V1PodSpec(
                    containers=[container],
                    volumes=volumes,
                ),
            ),
        )

    meta_kwargs = {"name": TEST_NAME, "namespace": TEST_NAME}
    if labels is not None:
        meta_kwargs["labels"] = labels
    if annotations is not None:
        meta_kwargs["annotations"] = annotations

    pytorchjob = KubeflowOrgV1PyTorchJob(
        api_version=constants.API_VERSION,
        kind=constants.PYTORCHJOB_KIND,
        metadata=V1ObjectMeta(**meta_kwargs),
        spec=KubeflowOrgV1PyTorchJobSpec(
            run_policy=KubeflowOrgV1RunPolicy(clean_pod_policy=None),
            pytorch_replica_specs=pytorch_replica_specs,
        ),
    )

    return pytorchjob


# Check if actual string contains all elements from the expected list.
class AnyStringWithElementsFromList:
    def __init__(self, expected):
        self.expected = expected

    def __eq__(self, actual):
        return all(e in str(actual) for e in self.expected)


def create_job_from_func(num_workers, packages_to_install=None, pip_index_url=None):

    command = constants.DEFAULT_COMMAND
    if num_workers > 1:
        args = [f'{constants.ENTRYPOINT_TORCH} "$program_path/ephemeral_script.py"']
    else:
        args = [f'{constants.ENTRYPOINT_PYTHON} "$program_path/ephemeral_script.py"']

    if pip_index_url and packages_to_install:
        args += [f"--index-url {pip_index_url} {packages_to_install[0]}"]

    job = create_job(command, AnyStringWithElementsFromList(args), num_workers)

    return job


def generate_job_with_status(
    job: constants.JOB_MODELS_TYPE,
    condition_type: str = constants.JOB_CONDITION_SUCCEEDED,
) -> constants.JOB_MODELS_TYPE:
    job.status = KubeflowOrgV1JobStatus(
        conditions=[
            KubeflowOrgV1JobCondition(
                type=condition_type,
                status=constants.CONDITION_STATUS_TRUE,
            )
        ]
    )
    return job


class DummyJobClass:
    def __init__(self, kind) -> None:
        self.kind = kind


def generate_pod(status, name=DUMMY_POD_NAME, timestamp=None):
    pod = Mock(metadata=Mock())
    pod.metadata.name = name
    pod.metadata.creation_timestamp = timestamp
    pod.status = status
    return pod


def mock_get_job_pods(*args, **kwargs):
    """Mock get_job_pods to return controlled pod objects"""
    namespace = kwargs.get("namespace")
    if namespace == f"pod {TIMEOUT}":
        raise TimeoutError()
    if namespace == f"pod {RUNTIME}":
        raise RuntimeError()
    if namespace == INVALID:
        raise ValueError()

    # Handle different test scenarios
    if namespace == MULTI_PODS:
        return [generate_pod(Mock(phase=RUNNING), f"pod-{i}") for i in range(3)]

    # To find relevant events, the pod's creation time must precede the event's creation time
    pod_creation_timestamp = EVENT_CREATION_TIMESTAMP - timedelta(seconds=1)
    pod = generate_pod(None, timestamp=pod_creation_timestamp)
    if namespace == PENDING_POD:
        pod.status = Mock(phase=constants.POD_PHASE_PENDING)
    elif namespace == NO_STATUS_POD:
        pod.status = None
    else:
        pod.status = Mock(phase=RUNNING)
    return [pod]


def mock_get_job(*args, **kwargs):
    """Mock get_job_pods to return controlled pod objects"""
    namespace = kwargs.get("namespace")
    if namespace == f"job {TIMEOUT}":
        raise TimeoutError()
    if namespace == f"job {RUNTIME}":
        raise RuntimeError()

    # Handle different test scenarios
    job = Mock()
    # To find relevant events, the job's creation time must precede the event's creation time
    job.metadata = Mock(
        creation_timestamp=EVENT_CREATION_TIMESTAMP - timedelta(seconds=1)
    )
    return job


def mock_read_namespaced_pod_log(*args, **kwargs):
    """Mock for reading pod logs"""
    if kwargs.get("namespace") == FAIL_LOGS:
        raise Exception("Failed to read logs")
    return "test log content"


def mock_watch(self, *args, **kwargs):
    namespace = kwargs.get("namespace")
    if namespace == FAIL_LOGS:
        raise Exception("Failed to read logs")
    if namespace == QUEUE_TIMEOUT:
        log_lines = [TIMEOUT]
    elif namespace == QUEUE_EMPTY:
        log_lines = [QUEUE_EMPTY]
    else:
        log_lines = ["line 1 of pod logs", "line 2 of pod logs", "line 3 of pod logs"]
    return iter(log_lines)


def mock_get_log_queue_pool(log_streams):
    mock_logs = []
    for stream in log_streams:
        # Convert iterator to list to preserve values
        log_lines = list(stream)
        mock_queue = Mock()
        # Use a list to maintain state between calls
        remaining_logs = log_lines.copy()  # Make a copy to avoid modifying original

        def get_next(timeout, logs=remaining_logs):
            if logs:
                log = logs.pop(0)
                if log == TIMEOUT:
                    raise TimeoutError
                if log == QUEUE_EMPTY:
                    raise queue.Empty
                return log
            return None

        mock_queue.get = Mock(side_effect=get_next)
        mock_queue.put = Mock()
        mock_logs.append(mock_queue)
    return mock_logs


def mock_list_namespaced_event(*args, **kwargs):
    """Mock for listing namespace events"""

    class MockEvent:
        def __init__(self, kind, name):
            self.involved_object = Mock(kind=kind)
            self.involved_object.name = name
            self.metadata = Mock(creation_timestamp=EVENT_CREATION_TIMESTAMP)
            self.message = f"{kind} Event 1"

    class MockEventList:
        def __init__(self):
            self.items = [
                MockEvent(constants.POD_KIND, DUMMY_POD_NAME),
                MockEvent(constants.PYTORCHJOB_KIND, TEST_NAME),
            ]

    if kwargs.get("namespace") == FAIL_EVENTS:
        raise Exception("Failed to read events")
    return MockEventList()


test_data_create_job = [
    (
        "valid flow",
        {"job": create_job(), "namespace": TEST_NAME},
        SUCCESS,
        create_job(),
    ),
    (
        "valid flow to create multi-node job with torchrun",
        {
            "name": TEST_NAME,
            "namespace": TEST_NAME,
            "train_func": lambda: print("Test Training Function"),
            "base_image": TEST_IMAGE,
            "num_workers": 3,
            "packages_to_install": ["boto3==1.34.14"],
            "pip_index_url": "https://pypi.custom.com/simple",
        },
        SUCCESS,
        create_job_from_func(
            num_workers=3,
            packages_to_install=["boto3==1.34.1"],
            pip_index_url="https://pypi.custom.com/simple",
        ),
    ),
    (
        "valid flow to create job with 1 worker",
        {
            "name": TEST_NAME,
            "namespace": TEST_NAME,
            "train_func": lambda: print("Test Training Function"),
            "base_image": TEST_IMAGE,
            "num_workers": 1,
        },
        SUCCESS,
        create_job_from_func(num_workers=1),
    ),
    (
        "valid flow to create job using image",
        {
            "name": TEST_NAME,
            "namespace": TEST_NAME,
            "base_image": TEST_IMAGE,
            "num_workers": 2,
        },
        SUCCESS,
        create_job(num_workers=2),
    ),
    (
        "invalid extra parameter",
        {
            "job": create_job(),
            "namespace": TEST_NAME,
            "base_image": "test_image",
        },
        ValueError,
        None,
    ),
    (
        "invalid job kind",
        {"job_kind": "invalid_job_kind"},
        ValueError,
        None,
    ),
    (
        "job name missing with train function",
        {"train_func": lambda: "test train function"},
        ValueError,
        None,
    ),
    (
        "job name missing with base image",
        {"base_image": "test_image"},
        ValueError,
        None,
    ),
    (
        "uncallable train function",
        {
            "name": TEST_NAME,
            "train_func": "uncallable train function",
        },
        ValueError,
        None,
    ),
    (
        "invalid number of workers",
        {
            "name": TEST_NAME,
            "num_workers": 0,
        },
        ValueError,
        None,
    ),
    (
        "num_procs_per_worker is set for TFJob",
        {
            "name": TEST_NAME,
            "job_kind": constants.TFJOB_KIND,
            "num_procs_per_worker": 5,
            "base_image": "test_image",
        },
        ValueError,
        None,
    ),
    (
        "num_chief_replicas and num_ps_replicas is set for PyTorchJov",
        {
            "name": TEST_NAME,
            "num_chief_replicas": 1,
            "num_ps_replicas": 1,
            "base_image": "test_image",
        },
        ValueError,
        None,
    ),
    (
        "paddle job can't be created using function",
        {
            "name": TEST_NAME,
            "train_func": lambda: "test train function",
            "job_kind": constants.PADDLEJOB_KIND,
        },
        ValueError,
        None,
    ),
    (
        "invalid job object",
        {"job": DummyJobClass(constants.TFJOB_KIND)},
        ValueError,
        None,
    ),
    (
        "create_namespaced_custom_object timeout error",
        {"job": create_job(), "namespace": TIMEOUT},
        TimeoutError,
        None,
    ),
    (
        "create_namespaced_custom_object runtime error",
        {"job": create_job(), "namespace": RUNTIME},
        RuntimeError,
        None,
    ),
    (
        "valid flow with env_vars as dict",
        {
            "name": TEST_NAME,
            "namespace": TEST_NAME,
            "env_vars": {"ENV_VAR": "env_value"},
            "base_image": TEST_IMAGE,
            "num_workers": 1,
        },
        SUCCESS,
        create_job(env_vars={"ENV_VAR": "env_value"}, num_workers=1),
    ),
    (
        "valid flow with env_vars as list",
        {
            "name": TEST_NAME,
            "namespace": TEST_NAME,
            "env_vars": [V1EnvVar(name="ENV_VAR", value="env_value")],
            "base_image": TEST_IMAGE,
            "num_workers": 2,
        },
        SUCCESS,
        create_job(
            env_vars=[V1EnvVar(name="ENV_VAR", value="env_value")], num_workers=2
        ),
    ),
    (
        "create job with a volume and a volume mount",
        {
            "name": TEST_NAME,
            "namespace": TEST_NAME,
            "base_image": TEST_IMAGE,
            "num_workers": 1,
            "volumes": [V1Volume(name="vol")],
            "volume_mounts": [V1VolumeMount(name="vol", mount_path="/mnt")],
        },
        SUCCESS,
        create_job(
            num_workers=1,
            volumes=[V1Volume(name="vol")],
            volume_mounts=[V1VolumeMount(name="vol", mount_path="/mnt")],
        ),
    ),
    (
        "invalid number of volume mount",
        {
            "name": TEST_NAME,
            "namespace": TEST_NAME,
            "base_image": TEST_IMAGE,
            "num_workers": 1,
            "volumes": [V1Volume(name="vol")],
        },
        ValueError,
        None,
    ),
    (
        "valid flow with labels",
        {
            "name": TEST_NAME,
            "namespace": TEST_NAME,
            "base_image": TEST_IMAGE,
            "num_workers": 1,
            "labels": {"upstream": "kubeflow", "component": "training"},
        },
        SUCCESS,
        create_job(
            num_workers=1,
            labels={"upstream": "kubeflow", "component": "training"},
            annotations=None,
        ),
    ),
    (
        "valid flow with annotations",
        {
            "name": TEST_NAME,
            "namespace": TEST_NAME,
            "base_image": TEST_IMAGE,
            "num_workers": 1,
            "annotations": {"purpose": "unit-test", "env": "test"},
        },
        SUCCESS,
        create_job(
            num_workers=1,
            labels=None,
            annotations={"purpose": "unit-test", "env": "test"},
        ),
    ),
]

test_data_get_job_pods = [
    (
        "valid flow with default namespace and default timeout",
        {
            "name": TEST_NAME,
        },
        f"{constants.JOB_NAME_LABEL}={TEST_NAME}",
        LIST_RESPONSE,
    ),
    (
        "invalid replica_type",
        {"name": TEST_NAME, "replica_type": "invalid_replica_type"},
        "Label not relevant",
        ValueError,
    ),
    (
        "invalid replica_type (uppercase)",
        {"name": TEST_NAME, "replica_type": constants.REPLICA_TYPE_WORKER},
        "Label not relevant",
        ValueError,
    ),
    (
        "valid flow with specific timeout, replica_index, replica_type and master role",
        {
            "name": TEST_NAME,
            "namespace": "test_namespace",
            "timeout": 60,
            "is_master": True,
            "replica_type": constants.REPLICA_TYPE_MASTER.lower(),
            "replica_index": 0,
        },
        f"{constants.JOB_NAME_LABEL}={TEST_NAME},"
        f"{constants.JOB_ROLE_LABEL}={constants.JOB_ROLE_MASTER},"
        f"{constants.REPLICA_TYPE_LABEL}={constants.REPLICA_TYPE_MASTER.lower()},"
        f"{constants.REPLICA_INDEX_LABEL}=0",
        LIST_RESPONSE,
    ),
    (
        "invalid flow with TimeoutError",
        {
            "name": TEST_NAME,
            "namespace": TIMEOUT,
        },
        "Label not relevant",
        TimeoutError,
    ),
    (
        "invalid flow with RuntimeError",
        {
            "name": TEST_NAME,
            "namespace": RUNTIME,
        },
        "Label not relevant",
        RuntimeError,
    ),
]

test_data_wait_for_job_conditions = [
    (
        "timeout waiting for succeeded condition",
        {
            "name": TEST_NAME,
            "namespace": TIMEOUT,
            "wait_timeout": 0,
        },
        TimeoutError,
    ),
    (
        "invalid expected condition",
        {
            "name": TEST_NAME,
            "namespace": "value",
            "expected_conditions": {"invalid"},
        },
        ValueError,
    ),
    (
        "invalid expected condition(lowercase)",
        {
            "name": TEST_NAME,
            "namespace": "value",
            "expected_conditions": {"succeeded"},
        },
        ValueError,
    ),
    (
        "job failed unexpectedly",
        {
            "name": TEST_NAME,
            "namespace": RUNTIME,
        },
        RuntimeError,
    ),
    (
        "valid case",
        {
            "name": TEST_NAME,
            "namespace": "test-namespace",
        },
        generate_job_with_status(create_job()),
    ),
    (
        "valid case with specified callback",
        {
            "name": TEST_NAME,
            "namespace": "test-namespace",
            "callback": lambda job: "test train function",
        },
        generate_job_with_status(create_job()),
    ),
]


test_data_get_job_pod_names = [
    (
        "valid flow",
        {
            "name": TEST_NAME,
            "namespace": MOCK_POD_OBJ,
        },
        [DUMMY_POD_NAME],
    ),
    (
        "valid flow with no pods available",
        {
            "name": TEST_NAME,
            "namespace": NO_PODS,
        },
        [],
    ),
]


test_data_update_job = [
    (
        "valid flow",
        {
            "name": TEST_NAME,
            "job": create_job(),
        },
        "No output",
    ),
    (
        "invalid job_kind",
        {
            "name": TEST_NAME,
            "job": create_job(),
            "job_kind": "invalid_job_kind",
        },
        ValueError,
    ),
    (
        "invalid flow with TimeoutError",
        {
            "name": TEST_NAME,
            "namespace": TIMEOUT,
            "job": create_job(),
        },
        TimeoutError,
    ),
    (
        "invalid flow with RuntimeError",
        {
            "name": TEST_NAME,
            "namespace": RUNTIME,
            "job": create_job(),
        },
        RuntimeError,
    ),
]

test_data_get_job = [
    (
        "valid flow with all parameters set",
        {
            "name": TEST_NAME,
            "namespace": TEST_NAME,
            "job_kind": constants.PYTORCHJOB_KIND,
            "timeout": 120,
        },
        SUCCESS,
    ),
    (
        "invalid flow with default namespace and a Job that doesn't exist",
        {"name": TEST_NAME, "job_kind": constants.TFJOB_KIND},
        RuntimeError,
    ),
    (
        "invalid flow incorrect parameter",
        {"name": TEST_NAME, "test": "example"},
        TypeError,
    ),
    (
        "invalid flow with incorrect value",
        {"name": TEST_NAME, "job_kind": INVALID},
        ValueError,
    ),
    (
        "runtime error case",
        {
            "name": TEST_NAME,
            "namespace": "runtime",
            "job_kind": constants.PYTORCHJOB_KIND,
        },
        RuntimeError,
    ),
    (
        "invalid flow with timeout error",
        {"name": TEST_NAME, "namespace": TIMEOUT},
        TimeoutError,
    ),
]

test_data_list_jobs = [
    (
        "valid flow with default namespace and default timeout",
        {},
        SUCCESS,
        [generate_job_with_status(create_job())],
    ),
    (
        "valid flow with all parameters set",
        {
            "namespace": TEST_NAME,
            "job_kind": constants.PYTORCHJOB_KIND,
            "timeout": 120,
        },
        SUCCESS,
        [generate_job_with_status(create_job())],
    ),
    (
        "valid flow with empty job list",
        {
            "namespace": "empty-namespace",
        },
        SUCCESS,
        [],
    ),
    (
        "valid flow with multiple jobs",
        {
            "namespace": "multi-jobs",
        },
        SUCCESS,
        [
            generate_job_with_status(create_job()),
            generate_job_with_status(create_job()),
        ],
    ),
    (
        "invalid flow with default namespace and a Job that doesn't exist",
        {"job_kind": constants.TFJOB_KIND},
        RuntimeError,
        None,
    ),
    (
        "invalid flow with incorrect parameter",
        {"test": "example"},
        TypeError,
        None,
    ),
    (
        "invalid flow with incorrect job_kind value",
        {"job_kind": "FailJob"},
        ValueError,
        None,
    ),
    (
        "runtime error case",
        {"namespace": RUNTIME},
        RuntimeError,
        None,
    ),
    (
        "timeout error case",
        {"namespace": TIMEOUT},
        TimeoutError,
        None,
    ),
]


test_data_delete_job = [
    (
        "valid flow with default namespace",
        {
            "name": TEST_NAME,
        },
        SUCCESS,
    ),
    (
        "invalid extra parameter",
        {"name": TEST_NAME, "namespace": TEST_NAME, "example": "test"},
        TypeError,
    ),
    (
        "invalid job kind",
        {"name": TEST_NAME, "job_kind": "invalid_job_kind"},
        RuntimeError,
    ),
    (
        "job name missing",
        {"namespace": TEST_NAME, "job_kind": constants.PYTORCHJOB_KIND},
        TypeError,
    ),
    (
        "delete_namespaced_custom_object timeout error",
        {"name": TEST_NAME, "namespace": TIMEOUT},
        TimeoutError,
    ),
    (
        "delete_namespaced_custom_object runtime error",
        {"name": TEST_NAME, "namespace": RUNTIME},
        RuntimeError,
    ),
    (
        "valid flow",
        {
            "name": TEST_NAME,
            "namespace": TEST_NAME,
            "job_kind": constants.PYTORCHJOB_KIND,
        },
        SUCCESS,
    ),
    (
        "valid flow with delete options",
        {
            "name": TEST_NAME,
            "delete_options": V1DeleteOptions(grace_period_seconds=30),
        },
        SUCCESS,
    ),
]


test_data_get_job_conditions = [
    (
        "valid flow with failed job condition",
        {"name": TEST_NAME, "namespace": FAILED},
        generate_job_with_status(
            create_job(), condition_type=constants.JOB_CONDITION_FAILED
        ),
    ),
    (
        "valid flow with restarting job condition",
        {"name": TEST_NAME, "namespace": RESTARTING},
        generate_job_with_status(
            create_job(), condition_type=constants.JOB_CONDITION_RESTARTING
        ),
    ),
    (
        "valid flow with running job condition",
        {"name": TEST_NAME, "namespace": RUNNING},
        generate_job_with_status(
            create_job(), condition_type=constants.JOB_CONDITION_RUNNING
        ),
    ),
    (
        "valid flow with created job condition",
        {"name": TEST_NAME, "namespace": CREATED},
        generate_job_with_status(
            create_job(), condition_type=constants.JOB_CONDITION_CREATED
        ),
    ),
    (
        "valid flow with all parameters set",
        {
            "name": TEST_NAME,
            "namespace": TEST_NAME,
            "job": create_job(),
            "job_kind": constants.PYTORCHJOB_KIND,
            "timeout": 120,
        },
        generate_job_with_status(create_job()),
    ),
    (
        "invalid flow with default namespace and a Job that doesn't exist",
        {"name": TEST_NAME, "job_kind": constants.TFJOB_KIND},
        RuntimeError,
    ),
    (
        "invalid flow incorrect parameter",
        {"name": TEST_NAME, "test": "example"},
        TypeError,
    ),
    (
        "invalid flow with incorrect value",
        {"name": TEST_NAME, "job_kind": INVALID},
        ValueError,
    ),
    (
        "runtime error case",
        {
            "name": TEST_NAME,
            "namespace": "runtime",
            "job_kind": constants.PYTORCHJOB_KIND,
        },
        RuntimeError,
    ),
    (
        "invalid flow with timeout error",
        {"name": TEST_NAME, "namespace": TIMEOUT},
        TimeoutError,
    ),
]


test_data_is_job_created = [
    (
        "valid flow with all parameters set",
        {
            "name": TEST_NAME,
            "namespace": CREATED,
            "job": create_job(),
            "job_kind": constants.PYTORCHJOB_KIND,
            "timeout": 120,
        },
        True,
    ),
    (
        "invalid flow with default namespace and a Job that doesn't exist",
        {"name": TEST_NAME, "job_kind": constants.TFJOB_KIND},
        RuntimeError,
    ),
    (
        "invalid flow incorrect parameter",
        {"name": TEST_NAME, "test": "example"},
        TypeError,
    ),
    (
        "invalid flow with incorrect value",
        {"name": TEST_NAME, "job_kind": INVALID},
        ValueError,
    ),
    (
        "runtime error case",
        {
            "name": TEST_NAME,
            "namespace": "runtime",
            "job_kind": constants.PYTORCHJOB_KIND,
        },
        RuntimeError,
    ),
    (
        "invalid flow with timeout error",
        {"name": TEST_NAME, "namespace": TIMEOUT},
        TimeoutError,
    ),
]


test_data_is_job_running = [
    (
        "valid flow with job that is running",
        {"name": TEST_NAME, "namespace": RUNNING},
        True,
    ),
    (
        "valid flow with all parameters set",
        {
            "name": TEST_NAME,
            "namespace": RUNNING,
            "job": create_job(),
            "job_kind": constants.PYTORCHJOB_KIND,
            "timeout": 120,
        },
        True,
    ),
    (
        "invalid flow with default namespace and a Job that doesn't exist",
        {"name": TEST_NAME, "job_kind": constants.TFJOB_KIND},
        RuntimeError,
    ),
    (
        "invalid flow incorrect parameter",
        {"name": TEST_NAME, "test": "example"},
        TypeError,
    ),
    (
        "invalid flow with incorrect value",
        {"name": TEST_NAME, "job_kind": INVALID},
        ValueError,
    ),
    (
        "runtime error case",
        {
            "name": TEST_NAME,
            "namespace": "runtime",
            "job_kind": constants.PYTORCHJOB_KIND,
        },
        RuntimeError,
    ),
    (
        "invalid flow with timeout error",
        {"name": TEST_NAME, "namespace": TIMEOUT},
        TimeoutError,
    ),
]


test_data_is_job_restarting = [
    (
        "valid flow with job that is restarting",
        {"name": TEST_NAME, "namespace": RESTARTING},
        True,
    ),
    (
        "valid flow with all parameters set",
        {
            "name": TEST_NAME,
            "namespace": RESTARTING,
            "job": create_job(),
            "job_kind": constants.PYTORCHJOB_KIND,
            "timeout": 120,
        },
        True,
    ),
    (
        "invalid flow with default namespace and a Job that doesn't exist",
        {"name": TEST_NAME, "job_kind": constants.TFJOB_KIND},
        RuntimeError,
    ),
    (
        "invalid flow incorrect parameter",
        {"name": TEST_NAME, "test": "example"},
        TypeError,
    ),
    (
        "invalid flow with incorrect value",
        {"name": TEST_NAME, "job_kind": INVALID},
        ValueError,
    ),
    (
        "runtime error case",
        {
            "name": TEST_NAME,
            "namespace": "runtime",
            "job_kind": constants.PYTORCHJOB_KIND,
        },
        RuntimeError,
    ),
    (
        "invalid flow with timeout error",
        {"name": TEST_NAME, "namespace": TIMEOUT},
        TimeoutError,
    ),
]


test_data_is_job_failed = [
    (
        "valid flow with job that is failed",
        {"name": TEST_NAME, "namespace": FAILED},
        True,
    ),
    (
        "valid flow with all parameters set",
        {
            "name": TEST_NAME,
            "namespace": FAILED,
            "job": create_job(),
            "job_kind": constants.PYTORCHJOB_KIND,
            "timeout": 120,
        },
        True,
    ),
    (
        "invalid flow with default namespace and a Job that doesn't exist",
        {"name": TEST_NAME, "job_kind": constants.TFJOB_KIND},
        RuntimeError,
    ),
    (
        "invalid flow incorrect parameter",
        {"name": TEST_NAME, "test": "example"},
        TypeError,
    ),
    (
        "invalid flow with incorrect value",
        {"name": TEST_NAME, "job_kind": INVALID},
        ValueError,
    ),
    (
        "runtime error case",
        {
            "name": TEST_NAME,
            "namespace": "runtime",
            "job_kind": constants.PYTORCHJOB_KIND,
        },
        RuntimeError,
    ),
    (
        "invalid flow with timeout error",
        {"name": TEST_NAME, "namespace": TIMEOUT},
        TimeoutError,
    ),
]


test_data_is_job_succeded = [
    (
        "valid flow with all parameters set",
        {
            "name": TEST_NAME,
            "namespace": SUCCEEDED,
            "job": create_job(),
            "job_kind": constants.PYTORCHJOB_KIND,
            "timeout": 120,
        },
        True,
    ),
    (
        "invalid flow with default namespace and a Job that doesn't exist",
        {"name": TEST_NAME, "job_kind": constants.TFJOB_KIND},
        RuntimeError,
    ),
    (
        "invalid flow incorrect parameter",
        {"name": TEST_NAME, "test": "example"},
        TypeError,
    ),
    (
        "invalid flow with incorrect value",
        {"name": TEST_NAME, "job_kind": INVALID},
        ValueError,
    ),
    (
        "runtime error case",
        {
            "name": TEST_NAME,
            "namespace": "runtime",
            "job_kind": constants.PYTORCHJOB_KIND,
        },
        RuntimeError,
    ),
    (
        "invalid flow with timeout error",
        {"name": TEST_NAME, "namespace": TIMEOUT},
        TimeoutError,
    ),
]

test_data_get_job_logs = [
    # Basic cases
    (
        "valid flow with default parameters",
        {
            "name": TEST_NAME,
        },
        {DUMMY_POD_NAME: "test log content"},
        {},
        SUCCESS,
    ),
    (
        "pod with pending status",
        {
            "name": TEST_NAME,
            "namespace": PENDING_POD,
        },
        {},  # No logs expected for pending pods
        {},
        SUCCESS,
    ),
    (
        "pod with pending status and follow",
        {
            "name": TEST_NAME,
            "namespace": PENDING_POD,
            "follow": True,
        },
        {},  # No logs expected for pending pods
        {},
        SUCCESS,
    ),
    (
        "pod with no status",
        {
            "name": TEST_NAME,
            "namespace": NO_STATUS_POD,
        },
        {},  # No logs expected
        {},
        SUCCESS,
    ),
    (
        "pod with no status and follow",
        {
            "name": TEST_NAME,
            "namespace": NO_STATUS_POD,
            "follow": True,
        },
        {},  # No logs expected
        {},
        SUCCESS,
    ),
    (
        "valid flow with logs and verbose",
        {
            "name": TEST_NAME,
            "namespace": TEST_NAME,
            "verbose": True,
        },
        {DUMMY_POD_NAME: "test log content"},
        {
            f"{constants.PYTORCHJOB_KIND.lower()}/{TEST_NAME}": [
                f"{EVENT_CREATION_TIMESTAMP.strftime('%Y-%m-%d %H:%M:%S')} PyTorchJob Event 1"
            ],
            f"{constants.POD_KIND.lower()}/{DUMMY_POD_NAME}": [
                f"{EVENT_CREATION_TIMESTAMP.strftime('%Y-%m-%d %H:%M:%S')} Pod Event 1"
            ],
        },
        SUCCESS,
    ),
    (
        "valid flow with worker logs",
        {
            "name": TEST_NAME,
            "namespace": TEST_NAME,
            "is_master": False,
            "replica_type": constants.REPLICA_TYPE_WORKER.lower(),
            "replica_index": 0,
        },
        {DUMMY_POD_NAME: "test log content"},
        {},
        SUCCESS,
    ),
    # Streaming cases
    (
        "valid flow with follow logs",
        {
            "name": TEST_NAME,
            "follow": True,
        },
        {
            DUMMY_POD_NAME: (
                "line 1 of pod logs" "line 2 of pod logs" "line 3 of pod logs"
            )
        },
        {},
        SUCCESS,
    ),
    (
        "valid flow with follow logs and multiple pods",
        {
            "name": TEST_NAME,
            "namespace": MULTI_PODS,
            "follow": True,
        },
        {
            "pod-0": ("line 1 of pod logs" "line 2 of pod logs" "line 3 of pod logs"),
            "pod-1": ("line 1 of pod logs" "line 2 of pod logs" "line 3 of pod logs"),
            "pod-2": ("line 1 of pod logs" "line 2 of pod logs" "line 3 of pod logs"),
        },
        {},
        SUCCESS,
    ),
    (
        "follow logs with queue empty",
        {
            "name": TEST_NAME,
            "namespace": QUEUE_EMPTY,
            "follow": True,
        },
        {},
        {},
        SUCCESS,
    ),
    # Error cases
    (
        "invalid replica type",
        {
            "name": TEST_NAME,
            "namespace": INVALID,
            "replica_type": "invalid_replica",
        },
        None,
        None,
        ValueError,
    ),
    (
        "timeout error when getting pods",
        {
            "name": TEST_NAME,
            "namespace": f"pod {TIMEOUT}",
        },
        None,
        None,
        TimeoutError,
    ),
    (
        "runtime error when getting pods",
        {
            "name": TEST_NAME,
            "namespace": f"pod {RUNTIME}",
        },
        None,
        None,
        RuntimeError,
    ),
    (
        "exception when reading logs with follow",
        {
            "name": TEST_NAME,
            "namespace": FAIL_LOGS,
            "follow": True,
        },
        None,
        None,
        Exception,
    ),
    (
        "runtime error when reading logs",
        {
            "name": TEST_NAME,
            "namespace": FAIL_LOGS,
        },
        None,
        None,
        RuntimeError,
    ),
    (
        "exception when reading events",
        {
            "name": TEST_NAME,
            "namespace": FAIL_EVENTS,
            "verbose": True,
        },
        None,
        None,
        Exception,
    ),
    (
        "timeout error when getting job",
        {
            "name": TEST_NAME,
            "namespace": f"job {TIMEOUT}",
            "verbose": True,
        },
        None,
        None,
        TimeoutError,
    ),
    (
        "runtime error when getting job",
        {
            "name": TEST_NAME,
            "namespace": f"job {RUNTIME}",
            "verbose": True,
        },
        None,
        None,
        RuntimeError,
    ),
    (
        "follow logs with queue timeout",
        {
            "name": TEST_NAME,
            "namespace": QUEUE_TIMEOUT,
            "follow": True,
        },
        {},  # Empty logs due to timeout
        {},
        TimeoutError,
    ),
]


@pytest.fixture
def training_client(request):
    mock_get_job_and_job_pods = (
        request.param.get("mock_get_job_and_job_pods", False)
        if hasattr(request, "param")
        else False
    )
    with patch(
        "kubernetes.client.CustomObjectsApi",
        return_value=Mock(
            create_namespaced_custom_object=Mock(side_effect=conditional_error_handler),
            patch_namespaced_custom_object=Mock(side_effect=conditional_error_handler),
            delete_namespaced_custom_object=Mock(side_effect=conditional_error_handler),
            get_namespaced_custom_object=Mock(
                side_effect=get_namespaced_custom_object_response
            ),
            list_namespaced_custom_object=Mock(
                side_effect=list_namespaced_custom_object_response
            ),
        ),
    ), patch(
        "kubernetes.client.CoreV1Api",
        return_value=Mock(
            list_namespaced_pod=Mock(side_effect=list_namespaced_pod_response),
            read_namespaced_pod_log=Mock(side_effect=mock_read_namespaced_pod_log),
            list_namespaced_event=Mock(side_effect=mock_list_namespaced_event),
        ),
    ), patch(
        "kubernetes.config.load_kube_config", return_value=Mock()
    ), patch(
        "kubernetes.watch.Watch",
        return_value=Mock(
            stream=Mock(side_effect=mock_watch),
        ),
    ), patch(
        "kubeflow.training.utils.utils.get_log_queue_pool",
        side_effect=mock_get_log_queue_pool,
    ):
        client = TrainingClient(job_kind=constants.PYTORCHJOB_KIND)
        if mock_get_job_and_job_pods:
            client.get_job_pods = Mock(side_effect=mock_get_job_pods)
            client.get_job = Mock(side_effect=mock_get_job)
        yield client


@pytest.mark.parametrize(
    "test_name,kwargs,expected_output,expected_job", test_data_create_job
)
def test_create_job(training_client, test_name, kwargs, expected_output, expected_job):
    """
    test create_job function of training client
    """
    print("Executing test:", test_name)
    try:
        training_client.create_job(**kwargs)

        assert expected_output == SUCCESS

        training_client.custom_api.create_namespaced_custom_object.assert_called_with(
            constants.GROUP,
            constants.VERSION,
            kwargs["namespace"],
            constants.JOB_PARAMETERS[constants.PYTORCHJOB_KIND]["plural"],
            expected_job,
        )
    except Exception as e:
        assert type(e) is expected_output
    print("test execution complete")


@pytest.mark.parametrize(
    "test_name,kwargs,expected_label_selector,expected_output",
    test_data_get_job_pods,
)
def test_get_job_pods(
    training_client, test_name, kwargs, expected_label_selector, expected_output
):
    """
    test get_job_pods function of training client
    """
    print("Executing test:", test_name)
    try:
        out = training_client.get_job_pods(**kwargs)
        # Verify that list_namespaced_pod called with specified arguments
        training_client.core_api.list_namespaced_pod.assert_called_with(
            kwargs.get("namespace", constants.DEFAULT_NAMESPACE),
            label_selector=expected_label_selector,
            async_req=True,
        )
        assert out[0].pop(TIMEOUT) == kwargs.get(TIMEOUT, constants.DEFAULT_TIMEOUT)
        assert out == expected_output
    except Exception as e:
        assert type(e) is expected_output
    print("test execution complete")


@pytest.mark.parametrize(
    "test_name,kwargs,expected_output",
    test_data_get_job_pod_names,
)
def test_get_job_pod_names(training_client, test_name, kwargs, expected_output):
    """
    test get_job_pod_names function of training client
    """
    print("Executing test:", test_name)
    out = training_client.get_job_pod_names(**kwargs)
    assert out == expected_output
    print("test execution complete")


@pytest.mark.parametrize("test_name,kwargs,expected_output", test_data_update_job)
def test_update_job(training_client, test_name, kwargs, expected_output):
    """
    test update_job function of training client
    """
    print("Executing test:", test_name)
    try:
        training_client.update_job(**kwargs)
        training_client.custom_api.patch_namespaced_custom_object.assert_called_with(
            constants.GROUP,
            constants.VERSION,
            kwargs.get("namespace", constants.DEFAULT_NAMESPACE),
            constants.JOB_PARAMETERS[kwargs.get("job_kind", training_client.job_kind)][
                "plural"
            ],
            kwargs.get("name"),
            kwargs.get("job"),
        )
    except Exception as e:
        assert type(e) is expected_output
    print("test execution complete")


@pytest.mark.parametrize(
    "test_name,kwargs,expected_output", test_data_wait_for_job_conditions
)
def test_wait_for_job_conditions(training_client, test_name, kwargs, expected_output):
    """
    test wait_for_job_conditions function of training client
    """
    print("Executing test:", test_name)
    try:
        out = training_client.wait_for_job_conditions(**kwargs)
        assert out == expected_output
    except Exception as e:
        assert type(e) is expected_output
    print("test execution complete")


@pytest.mark.parametrize("test_name,kwargs,expected_output", test_data_delete_job)
def test_delete_job(training_client, test_name, kwargs, expected_output):
    """
    test delete_job function of training client
    """
    print("Executing test: ", test_name)
    try:
        training_client.delete_job(**kwargs)
        assert expected_output == SUCCESS
    except Exception as e:
        assert type(e) is expected_output

    print("test execution complete")


@pytest.mark.parametrize("test_name,kwargs,expected_output", test_data_get_job)
def test_get_job(training_client, test_name, kwargs, expected_output):
    """
    test get_job function of training client
    """
    print("Executing test: ", test_name)

    try:
        training_client.get_job(**kwargs)
        assert expected_output == SUCCESS
    except Exception as e:
        assert type(e) is expected_output

    print("test execution complete")


@pytest.mark.parametrize(
    "test_name,kwargs,expected_output", test_data_get_job_conditions
)
def test_get_job_conditions(training_client, test_name, kwargs, expected_output):
    """
    test get_job_conditions function of training client
    """
    print("Executing test: ", test_name)

    try:
        training_client.get_job_conditions(**kwargs)
        if kwargs.get("namespace") is TEST_NAME:
            assert expected_output == generate_job_with_status(create_job())
        else:
            assert expected_output == generate_job_with_status(
                create_job(), condition_type=kwargs.get("namespace")
            )
    except Exception as e:
        assert type(e) is expected_output

    print("test execution complete")


@pytest.mark.parametrize("test_name,kwargs,expected_output", test_data_is_job_created)
def test_is_job_created(training_client, test_name, kwargs, expected_output):
    """
    test is_job_created function of training client
    """
    print("Executing test: ", test_name)

    try:
        training_client.is_job_created(**kwargs)
        if kwargs.get("namespace") is not (CREATED or RUNTIME or TIMEOUT):
            assert expected_output is False
        else:
            assert expected_output is True
    except Exception as e:
        assert type(e) is expected_output

    print("test execution complete")


@pytest.mark.parametrize("test_name,kwargs,expected_output", test_data_is_job_running)
def test_is_job_running(training_client, test_name, kwargs, expected_output):
    """
    test is_job_running function of training client
    """
    print("Executing test: ", test_name)

    try:
        training_client.is_job_running(**kwargs)
        if kwargs.get("namespace") is not (RUNNING or RUNTIME or TIMEOUT):
            assert expected_output is False
        else:
            assert expected_output is True
    except Exception as e:
        assert type(e) is expected_output

    print("test execution complete")


@pytest.mark.parametrize(
    "test_name,kwargs,expected_output", test_data_is_job_restarting
)
def test_is_job_restarting(training_client, test_name, kwargs, expected_output):
    """
    test is_is_job_restarting function of training client
    """
    print("Executing test: ", test_name)

    try:
        training_client.is_job_restarting(**kwargs)
        if kwargs.get("namespace") is not (RESTARTING or RUNTIME or TIMEOUT):
            assert expected_output is False
        else:
            assert expected_output is True
    except Exception as e:
        assert type(e) is expected_output

    print("test execution complete")


@pytest.mark.parametrize("test_name,kwargs,expected_output", test_data_is_job_failed)
def test_is_job_failed(training_client, test_name, kwargs, expected_output):
    """
    test is_is_job_failed function of training client
    """
    print("Executing test: ", test_name)

    try:
        training_client.is_job_failed(**kwargs)
        if kwargs.get("namespace") is not (FAILED or RUNTIME or TIMEOUT):
            assert expected_output is False
        else:
            assert expected_output is True
    except Exception as e:
        assert type(e) is expected_output

    print("test execution complete")


@pytest.mark.parametrize("test_name,kwargs,expected_output", test_data_is_job_succeded)
def test_is_job_succeeded(training_client, test_name, kwargs, expected_output):
    """
    test is_job_succeeded function of training client
    """
    print("Executing test: ", test_name)

    try:
        training_client.is_job_succeeded(**kwargs)
        if kwargs.get("namespace") is not (SUCCEEDED or RUNTIME or TIMEOUT):
            assert expected_output is False
        else:
            assert expected_output is True
    except Exception as e:
        assert type(e) is expected_output

    print("test execution complete")


@pytest.mark.parametrize(
    "test_name,kwargs,expected_status,expected_response", test_data_list_jobs
)
def test_list_jobs(
    training_client, test_name, kwargs, expected_status, expected_response
):
    """
    test list_jobs function of training client
    """
    print("Executing test: ", test_name)
    try:
        resp = training_client.list_jobs(**kwargs)
        assert expected_status == SUCCESS
        assert isinstance(resp, list)
        assert len(resp) == len(expected_response)
        for actual_job, expected_job in zip(resp, expected_response):
            assert actual_job.to_dict() == expected_job.to_dict()
    except Exception as e:
        assert type(e) is expected_status

    print("test execution complete")


@pytest.mark.parametrize(
    "training_client", [{"mock_get_job_and_job_pods": True}], indirect=True
)
@pytest.mark.parametrize(
    "test_name,kwargs,expected_logs,expected_events,expected_output",
    test_data_get_job_logs,
)
def test_get_job_logs(
    training_client,
    test_name,
    kwargs,
    expected_logs,
    expected_events,
    expected_output,
):
    """
    test get_job_logs function of training client
    """
    print("Executing test:", test_name)

    try:
        logs_dict, events_dict = training_client.get_job_logs(**kwargs)

        assert expected_output == SUCCESS
        assert logs_dict == expected_logs

        if kwargs.get("verbose", False):
            assert events_dict == expected_events
        else:
            assert events_dict == {}

        # Verify API calls
        training_client.get_job_pods.assert_called_with(
            name=kwargs["name"],
            namespace=kwargs.get("namespace", constants.DEFAULT_NAMESPACE),
            is_master=kwargs.get("is_master", True),
            replica_type=kwargs.get("replica_type"),
            replica_index=kwargs.get("replica_index"),
            timeout=kwargs.get("timeout", constants.DEFAULT_TIMEOUT),
        )

        if kwargs.get("verbose", False):
            training_client.get_job.assert_called_with(
                name=kwargs["name"],
                namespace=kwargs.get("namespace", constants.DEFAULT_NAMESPACE),
            )
            training_client.core_api.list_namespaced_event.assert_called_with(
                namespace=kwargs.get("namespace", constants.DEFAULT_NAMESPACE)
            )

    except Exception as e:
        assert type(e) is expected_output

    print("test execution complete")



# Copyright 2023 kubeflow.org.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

TEST_GANG_SCHEDULER_NAME_ENV_KEY = "GANG_SCHEDULER_NAME"
TEST_GANG_SCHEDULER_NAME_SCHEDULER_PLUGINS = "scheduler-plugins"
TEST_GANG_SCHEDULER_NAME_VOLCANO = "volcano"
TEST_GANG_SCHEDULER_NAME_NONE = "none"

GANG_SCHEDULERS = {
    TEST_GANG_SCHEDULER_NAME_SCHEDULER_PLUGINS,
    TEST_GANG_SCHEDULER_NAME_VOLCANO,
}
NONE_GANG_SCHEDULERS = {TEST_GANG_SCHEDULER_NAME_NONE, ""}

DEFAULT_SCHEDULER_PLUGINS_NAME = "scheduler-plugins-scheduler"



# Copyright 2021 kubeflow.org.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import logging
import pytest
from typing import Optional

from kubernetes.client import V1PodTemplateSpec
from kubernetes.client import V1ObjectMeta
from kubernetes.client import V1PodSpec
from kubernetes.client import V1Container
from kubernetes.client import V1ResourceRequirements

from kubeflow.training import TrainingClient
from kubeflow.training import KubeflowOrgV1ReplicaSpec
from kubeflow.training import KubeflowOrgV1PyTorchJob
from kubeflow.training import KubeflowOrgV1PyTorchJobSpec
from kubeflow.training import KubeflowOrgV1RunPolicy
from kubeflow.training import KubeflowOrgV1SchedulingPolicy
from kubeflow.training import constants

import test.e2e.utils as utils
from test.e2e.constants import TEST_GANG_SCHEDULER_NAME_ENV_KEY
from test.e2e.constants import GANG_SCHEDULERS, NONE_GANG_SCHEDULERS

logging.basicConfig(format="%(message)s")
logging.getLogger("kubeflow.training.api.training_client").setLevel(logging.DEBUG)

TRAINING_CLIENT = TrainingClient(job_kind=constants.PYTORCHJOB_KIND)
CONTAINER_NAME = "pytorch"
GANG_SCHEDULER_NAME = os.getenv(TEST_GANG_SCHEDULER_NAME_ENV_KEY, "")


@pytest.mark.skipif(
    GANG_SCHEDULER_NAME in NONE_GANG_SCHEDULERS,
    reason="For gang-scheduling",
)
def test_sdk_e2e_with_gang_scheduling(job_namespace):
    JOB_NAME = "pytorchjob-gang-scheduling"
    container = generate_container()

    master = KubeflowOrgV1ReplicaSpec(
        replicas=1,
        restart_policy="OnFailure",
        template=V1PodTemplateSpec(
            metadata=V1ObjectMeta(
                annotations={constants.ISTIO_SIDECAR_INJECTION: "false"}
            ),
            spec=V1PodSpec(
                scheduler_name=utils.get_pod_spec_scheduler_name(GANG_SCHEDULER_NAME),
                containers=[container],
            ),
        ),
    )

    worker = KubeflowOrgV1ReplicaSpec(
        replicas=1,
        restart_policy="OnFailure",
        template=V1PodTemplateSpec(
            metadata=V1ObjectMeta(
                annotations={constants.ISTIO_SIDECAR_INJECTION: "false"}
            ),
            spec=V1PodSpec(
                scheduler_name=utils.get_pod_spec_scheduler_name(GANG_SCHEDULER_NAME),
                containers=[container],
            ),
        ),
    )

    unschedulable_pytorchjob = generate_pytorchjob(
        job_namespace,
        JOB_NAME,
        master,
        worker,
        KubeflowOrgV1SchedulingPolicy(min_available=10),
    )
    schedulable_pytorchjob = generate_pytorchjob(
        job_namespace,
        JOB_NAME,
        master,
        worker,
        KubeflowOrgV1SchedulingPolicy(min_available=2),
    )

    TRAINING_CLIENT.create_job(job=unschedulable_pytorchjob, namespace=job_namespace)
    logging.info(f"List of created {TRAINING_CLIENT.job_kind}s")
    logging.info(TRAINING_CLIENT.list_jobs(job_namespace))

    try:
        utils.verify_unschedulable_job_e2e(TRAINING_CLIENT, JOB_NAME, job_namespace)
    except Exception as e:
        utils.print_job_results(TRAINING_CLIENT, JOB_NAME, job_namespace)
        TRAINING_CLIENT.delete_job(JOB_NAME, job_namespace)
        raise Exception(f"PyTorchJob E2E fails. Exception: {e}")

    TRAINING_CLIENT.update_job(schedulable_pytorchjob, JOB_NAME, job_namespace)
    logging.info(f"List of updated {TRAINING_CLIENT.job_kind}s")
    logging.info(TRAINING_CLIENT.list_jobs(job_namespace))

    try:
        utils.verify_job_e2e(TRAINING_CLIENT, JOB_NAME, job_namespace, wait_timeout=900)
    except Exception as e:
        utils.print_job_results(TRAINING_CLIENT, JOB_NAME, job_namespace)
        TRAINING_CLIENT.delete_job(JOB_NAME, job_namespace)
        raise Exception(f"PyTorchJob E2E fails. Exception: {e}")

    utils.print_job_results(TRAINING_CLIENT, JOB_NAME, job_namespace)
    TRAINING_CLIENT.delete_job(JOB_NAME, job_namespace)


@pytest.mark.skipif(
    GANG_SCHEDULER_NAME in GANG_SCHEDULERS,
    reason="For plain scheduling",
)
def test_sdk_e2e(job_namespace):
    JOB_NAME = "pytorchjob-e2e"
    container = generate_container()

    master = KubeflowOrgV1ReplicaSpec(
        replicas=1,
        restart_policy="OnFailure",
        template=V1PodTemplateSpec(
            metadata=V1ObjectMeta(
                annotations={constants.ISTIO_SIDECAR_INJECTION: "false"}
            ),
            spec=V1PodSpec(containers=[container]),
        ),
    )

    worker = KubeflowOrgV1ReplicaSpec(
        replicas=1,
        restart_policy="OnFailure",
        template=V1PodTemplateSpec(
            metadata=V1ObjectMeta(
                annotations={constants.ISTIO_SIDECAR_INJECTION: "false"}
            ),
            spec=V1PodSpec(containers=[container]),
        ),
    )

    pytorchjob = generate_pytorchjob(job_namespace, JOB_NAME, master, worker)

    TRAINING_CLIENT.create_job(job=pytorchjob, namespace=job_namespace)
    logging.info(f"List of created {TRAINING_CLIENT.job_kind}s")
    logging.info(TRAINING_CLIENT.list_jobs(job_namespace))

    try:
        utils.verify_job_e2e(TRAINING_CLIENT, JOB_NAME, job_namespace, wait_timeout=900)
    except Exception as e:
        utils.print_job_results(TRAINING_CLIENT, JOB_NAME, job_namespace)
        TRAINING_CLIENT.delete_job(JOB_NAME, job_namespace)
        raise Exception(f"PyTorchJob E2E fails. Exception: {e}")

    utils.print_job_results(TRAINING_CLIENT, JOB_NAME, job_namespace)
    TRAINING_CLIENT.delete_job(JOB_NAME, job_namespace)


@pytest.mark.skipif(
    GANG_SCHEDULER_NAME in GANG_SCHEDULERS,
    reason="For plain scheduling",
)
def test_sdk_e2e_managed_by(job_namespace):
    JOB_NAME = "pytorchjob-e2e"
    container = generate_container()

    master = KubeflowOrgV1ReplicaSpec(
        replicas=1,
        restart_policy="OnFailure",
        template=V1PodTemplateSpec(
            metadata=V1ObjectMeta(
                annotations={constants.ISTIO_SIDECAR_INJECTION: "false"}
            ),
            spec=V1PodSpec(containers=[container]),
        ),
    )

    worker = KubeflowOrgV1ReplicaSpec(
        replicas=1,
        restart_policy="OnFailure",
        template=V1PodTemplateSpec(
            metadata=V1ObjectMeta(
                annotations={constants.ISTIO_SIDECAR_INJECTION: "false"}
            ),
            spec=V1PodSpec(containers=[container]),
        ),
    )

    #1. Job created with default value: 'kubeflow.org/training-operator' - job created and status updated
    #2. Job created with kueue value: 'kueue.x-k8s.io/multikueue' - job created but status not updated
    #3. Job created with invalid value (not acceptable by the webhook) - job not created
    controllers = {
        JOB_NAME+"-default-controller": 'kubeflow.org/training-operator',
        JOB_NAME+"-multikueue-controller": 'kueue.x-k8s.io/multikueue', 
        JOB_NAME+"-invalid-controller": 'kueue.x-k8s.io/other-controller',
    }
    for job_name, managed_by in controllers.items():
        pytorchjob = generate_pytorchjob(job_namespace, job_name, master, worker, managed_by=managed_by)
        try:
            TRAINING_CLIENT.create_job(job=pytorchjob, namespace=job_namespace)
        except Exception as e:
            if "invalid" in str(job_name):
                error_message = f"Failed to create PyTorchJob: {job_namespace}/{job_name}"
                assert error_message in str(e), f"Unexpected error: {e}"
            else:
                raise Exception(f"PyTorchJob E2E fails. Exception: {e}")
    
    logging.info(f"List of created {TRAINING_CLIENT.job_kind}s")
    jobs = TRAINING_CLIENT.list_jobs(job_namespace)
    logging.info(jobs)

    try:
        #Only jobs with valid controllers should be created, 2 out of 3 satisfy this condition: 'kubeflow.org/training-operator' and 'kueue.x-k8s.io/multikueue'
        if len(jobs) != 2:
            raise Exception(f"Too many PyTorchJobs created {jobs}")
    
        for job in jobs:
            if job._metadata.name == 'kubeflow.org/training-operator':
                utils.verify_job_e2e(TRAINING_CLIENT, job._metadata.name, job_namespace, wait_timeout=900)
            if job._metadata.name == 'kueue.x-k8s.io/multikueue':
                conditions = TRAINING_CLIENT.get_job_conditions(job._metadata.name, job_namespace, TRAINING_CLIENT.job_kind, job)
                if len(conditions) != 0:
                    raise Exception(f"{TRAINING_CLIENT.job_kind} conditions {conditions} should not be updated, externally managed by {managed_by}")


    except Exception as e:
        utils.print_job_results(TRAINING_CLIENT, JOB_NAME, job_namespace)
        TRAINING_CLIENT.delete_job(JOB_NAME, job_namespace)
        raise Exception(f"PyTorchJob E2E fails. Exception: {e}")

    for job in jobs:
        utils.print_job_results(TRAINING_CLIENT, job._metadata.name, job_namespace)
        TRAINING_CLIENT.delete_job(job._metadata.name, job_namespace)

@pytest.mark.skipif(
    GANG_SCHEDULER_NAME in GANG_SCHEDULERS,
    reason="For plain scheduling",
)
def test_sdk_e2e_create_from_func(job_namespace):
    JOB_NAME = "pytorchjob-from-func"

    def train_func():
        import time

        for i in range(10):
            print(f"Start training for Epoch {i}")
            time.sleep(1)

    num_workers = 3

    TRAINING_CLIENT.create_job(
        name=JOB_NAME,
        namespace=job_namespace,
        train_func=train_func,
        num_workers=num_workers,
    )

    logging.info(f"List of created {TRAINING_CLIENT.job_kind}s")
    logging.info(TRAINING_CLIENT.list_jobs(job_namespace))

    try:
        utils.verify_job_e2e(TRAINING_CLIENT, JOB_NAME, job_namespace, wait_timeout=900)
    except Exception as e:
        utils.print_job_results(TRAINING_CLIENT, JOB_NAME, job_namespace)
        TRAINING_CLIENT.delete_job(JOB_NAME, job_namespace)
        raise Exception(f"PyTorchJob create from function E2E fails. Exception: {e}")

    # Verify that PyTorchJob has correct pods.
    pod_names = TRAINING_CLIENT.get_job_pod_names(
        name=JOB_NAME, namespace=job_namespace
    )

    if len(pod_names) != num_workers or f"{JOB_NAME}-worker-0" not in pod_names:
        raise Exception(f"PyTorchJob has incorrect pods: {pod_names}")

    utils.print_job_results(TRAINING_CLIENT, JOB_NAME, job_namespace)
    TRAINING_CLIENT.delete_job(JOB_NAME, job_namespace)


@pytest.mark.skipif(
    GANG_SCHEDULER_NAME in GANG_SCHEDULERS,
    reason="For plain scheduling",
)
def test_sdk_e2e_create_from_image(job_namespace):
    JOB_NAME = "pytorchjob-from-image"

    TRAINING_CLIENT.create_job(
        name=JOB_NAME,
        namespace=job_namespace,
        base_image="docker.io/hello-world",
        num_workers=1,
    )

    logging.info(f"List of created {TRAINING_CLIENT.job_kind}s")
    logging.info(TRAINING_CLIENT.list_jobs(job_namespace))

    try:
        utils.verify_job_e2e(TRAINING_CLIENT, JOB_NAME, job_namespace, wait_timeout=900)
    except Exception as e:
        utils.print_job_results(TRAINING_CLIENT, JOB_NAME, job_namespace)
        TRAINING_CLIENT.delete_job(JOB_NAME, job_namespace)
        raise Exception(f"PyTorchJob create from function E2E fails. Exception: {e}")

    utils.print_job_results(TRAINING_CLIENT, JOB_NAME, job_namespace)
    TRAINING_CLIENT.delete_job(JOB_NAME, job_namespace)


def generate_pytorchjob(
    job_namespace: str,
    job_name: str,
    master: KubeflowOrgV1ReplicaSpec,
    worker: KubeflowOrgV1ReplicaSpec,
    scheduling_policy: Optional[KubeflowOrgV1SchedulingPolicy] = None,
    managed_by: Optional[str] = None,
) -> KubeflowOrgV1PyTorchJob:
    return KubeflowOrgV1PyTorchJob(
        api_version=constants.API_VERSION,
        kind=constants.PYTORCHJOB_KIND,
        metadata=V1ObjectMeta(name=job_name, namespace=job_namespace),
        spec=KubeflowOrgV1PyTorchJobSpec(
            run_policy=KubeflowOrgV1RunPolicy(
                clean_pod_policy="None",
                scheduling_policy=scheduling_policy,
                managed_by=managed_by,
            ),
            pytorch_replica_specs={"Master": master, "Worker": worker},
        ),
    )


def generate_container() -> V1Container:
    return V1Container(
        name=CONTAINER_NAME,
        image="kubeflow/pytorch-dist-mnist:latest",
        args=["--backend", "gloo", "--epochs", "1"],
        resources=V1ResourceRequirements(limits={"memory": "2Gi", "cpu": "0.8"}),
    )



# coding: utf-8

"""
    Kubeflow Training SDK

    Python SDK for Kubeflow Training  # noqa: E501

    The version of the OpenAPI document: v1.9.2
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from kubeflow.training.configuration import Configuration


class KubeflowOrgV1TFJobSpec(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'enable_dynamic_worker': 'bool',
        'run_policy': 'KubeflowOrgV1RunPolicy',
        'success_policy': 'str',
        'tf_replica_specs': 'dict(str, KubeflowOrgV1ReplicaSpec)'
    }

    attribute_map = {
        'enable_dynamic_worker': 'enableDynamicWorker',
        'run_policy': 'runPolicy',
        'success_policy': 'successPolicy',
        'tf_replica_specs': 'tfReplicaSpecs'
    }

    def __init__(self, enable_dynamic_worker=None, run_policy=None, success_policy=None, tf_replica_specs=None, local_vars_configuration=None):  # noqa: E501
        """KubeflowOrgV1TFJobSpec - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._enable_dynamic_worker = None
        self._run_policy = None
        self._success_policy = None
        self._tf_replica_specs = None
        self.discriminator = None

        if enable_dynamic_worker is not None:
            self.enable_dynamic_worker = enable_dynamic_worker
        self.run_policy = run_policy
        if success_policy is not None:
            self.success_policy = success_policy
        self.tf_replica_specs = tf_replica_specs

    @property
    def enable_dynamic_worker(self):
        """Gets the enable_dynamic_worker of this KubeflowOrgV1TFJobSpec.  # noqa: E501

        A switch to enable dynamic worker  # noqa: E501

        :return: The enable_dynamic_worker of this KubeflowOrgV1TFJobSpec.  # noqa: E501
        :rtype: bool
        """
        return self._enable_dynamic_worker

    @enable_dynamic_worker.setter
    def enable_dynamic_worker(self, enable_dynamic_worker):
        """Sets the enable_dynamic_worker of this KubeflowOrgV1TFJobSpec.

        A switch to enable dynamic worker  # noqa: E501

        :param enable_dynamic_worker: The enable_dynamic_worker of this KubeflowOrgV1TFJobSpec.  # noqa: E501
        :type: bool
        """

        self._enable_dynamic_worker = enable_dynamic_worker

    @property
    def run_policy(self):
        """Gets the run_policy of this KubeflowOrgV1TFJobSpec.  # noqa: E501


        :return: The run_policy of this KubeflowOrgV1TFJobSpec.  # noqa: E501
        :rtype: KubeflowOrgV1RunPolicy
        """
        return self._run_policy

    @run_policy.setter
    def run_policy(self, run_policy):
        """Sets the run_policy of this KubeflowOrgV1TFJobSpec.


        :param run_policy: The run_policy of this KubeflowOrgV1TFJobSpec.  # noqa: E501
        :type: KubeflowOrgV1RunPolicy
        """
        if self.local_vars_configuration.client_side_validation and run_policy is None:  # noqa: E501
            raise ValueError("Invalid value for `run_policy`, must not be `None`")  # noqa: E501

        self._run_policy = run_policy

    @property
    def success_policy(self):
        """Gets the success_policy of this KubeflowOrgV1TFJobSpec.  # noqa: E501

        SuccessPolicy defines the policy to mark the TFJob as succeeded. Default to \"\", using the default rules.  # noqa: E501

        :return: The success_policy of this KubeflowOrgV1TFJobSpec.  # noqa: E501
        :rtype: str
        """
        return self._success_policy

    @success_policy.setter
    def success_policy(self, success_policy):
        """Sets the success_policy of this KubeflowOrgV1TFJobSpec.

        SuccessPolicy defines the policy to mark the TFJob as succeeded. Default to \"\", using the default rules.  # noqa: E501

        :param success_policy: The success_policy of this KubeflowOrgV1TFJobSpec.  # noqa: E501
        :type: str
        """

        self._success_policy = success_policy

    @property
    def tf_replica_specs(self):
        """Gets the tf_replica_specs of this KubeflowOrgV1TFJobSpec.  # noqa: E501

        A map of TFReplicaType (type) to ReplicaSpec (value). Specifies the TF cluster configuration. For example,   {     \"PS\": ReplicaSpec,     \"Worker\": ReplicaSpec,   }  # noqa: E501

        :return: The tf_replica_specs of this KubeflowOrgV1TFJobSpec.  # noqa: E501
        :rtype: dict(str, KubeflowOrgV1ReplicaSpec)
        """
        return self._tf_replica_specs

    @tf_replica_specs.setter
    def tf_replica_specs(self, tf_replica_specs):
        """Sets the tf_replica_specs of this KubeflowOrgV1TFJobSpec.

        A map of TFReplicaType (type) to ReplicaSpec (value). Specifies the TF cluster configuration. For example,   {     \"PS\": ReplicaSpec,     \"Worker\": ReplicaSpec,   }  # noqa: E501

        :param tf_replica_specs: The tf_replica_specs of this KubeflowOrgV1TFJobSpec.  # noqa: E501
        :type: dict(str, KubeflowOrgV1ReplicaSpec)
        """
        if self.local_vars_configuration.client_side_validation and tf_replica_specs is None:  # noqa: E501
            raise ValueError("Invalid value for `tf_replica_specs`, must not be `None`")  # noqa: E501

        self._tf_replica_specs = tf_replica_specs

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, KubeflowOrgV1TFJobSpec):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, KubeflowOrgV1TFJobSpec):
            return True

        return self.to_dict() != other.to_dict()
